##############################################################
# This code is part of the MAterials Simulation Toolkit (MAST)
# 
# Maintainer: Tam Mayeshiba
# Last updated: 2014-05-12 by Zhewen Song
##############################################################
from pymatgen.io.vaspio import Poscar
from pymatgen.io.vaspio import Outcar
from pymatgen.io.vaspio import Potcar
from pymatgen.io.vaspio import Incar
from pymatgen.io.vaspio import Kpoints
from pymatgen.io.vaspio.vasp_output import Vasprun
from pymatgen.io.cifio import CifParser
from pymatgen.core.sites import PeriodicSite
from pymatgen.core.structure import Structure
from MAST.utility import dirutil
from MAST.utility.mastfile import MASTFile
from MAST.utility import MASTError
from MAST.utility.metadata import Metadata
from MAST.ingredients.pmgextend.structure_extensions import StructureExtensions
import os
import shutil
import logging
import subprocess
from MAST.ingredients.checker import BaseChecker
class VaspChecker(BaseChecker):
    """VASP checker functions
    """
    def __init__(self, **kwargs):
        allowed_keys = {
            'name' : (str, str(), 'Name of directory'),
            'program_keys': (dict, dict(), 'Dictionary of program keywords'),
            'structure': (Structure, None, 'Pymatgen Structure object')
            }
        BaseChecker.__init__(self, allowed_keys, **kwargs)
        self.metafile = Metadata(metafile='%s/metadata.txt' % self.keywords['name'])

    def get_structure_from_file(self, myfilepath=""):
        """Get the structure from a specified file path.
            For VASP, this is a POSCAR-type file.
            Args:
                myfilepath <str>: File path for structure.
        """
        return Poscar.from_file(myfilepath).structure

    def get_initial_structure_from_directory(self,mydir=""):
        """Get the initial structure.
            For VASP, this is the POSCAR file.
            Args:
                mydir <str>: Directory. If no directory is given,
                    use current ingredient directory given as
                    keyword 'name'
        """
        if mydir == "":
            mydir = self.keywords['name']
        return Poscar.from_file("%s/POSCAR" % mydir).structure
    
    def get_final_structure_from_directory(self, mydir=""):
        """Get the final structure.
            For VASP, this is the CONTCAR file.
            Args:
                mydir <str>: Directory. If no directory is given,
                    use current ingredient directory given
                    as keyword 'name'
        """
        if mydir == "":
            mydir = self.keywords['name']
        return Poscar.from_file("%s/CONTCAR" % mydir).structure
    
    def forward_final_structure_file(self, childpath, newname="POSCAR"):
        """Forward the final structure.
            For VASP, this is the CONTCAR.
            Args:
                childpath <str>: Path of child ingredient
                newname <str>: new name (default 'POSCAR')
        """
        return self.copy_a_file(childpath, "CONTCAR", newname)

    def forward_initial_structure_file(self, childpath, newname="POSCAR"):
        """Forward the initial structure.
            For VASP, this is the POSCAR. This function is
            used after phonon calculations, where the CONTCAR
            contains the last displacement. To forward to PHON,
            the POSCAR (without displacements) should be used.
            Args:
                childpath <str>: Path of child ingredient
                newname <str>: new name (default 'POSCAR')
        """
        return self.copy_a_file(childpath, "POSCAR", newname)

    def forward_dynamical_matrix_file(self, childpath, newname="DYNMAT"):
        """Forward the dynamical matrix.
            For VASP, this is the DYNMAT file.
            Args:
                childpath <str>: Path of child ingredient
                newname <str>: new name (default 'DYNMAT')
        """
        return self.copy_a_file(childpath, "DYNMAT", newname)

    def forward_displacement_file(self, childpath, newname="XDATCAR"):
        """Forward displacement information.
            For VASP, this is the XDATCAR file.
            Args:
                childpath <str>: Path of child ingredient
                newname <str>: new name (default 'XDATCAR')
        """
        return self.copy_a_file(childpath, "XDATCAR", newname)

    def forward_energy_file(self, childpath, newname="OSZICAR"):
        """Forward the energy file.
            For VASP, this is the OSZICAR file.
            Args:
                childpath <str>: Path of child ingredient
                newname <str>: new name (default 'OSZICAR')
        """
        return self.copy_a_file(childpath, "OSZICAR", newname)

    def is_frozen(self):
        """Check if single VASP non-NEB calculation is frozen.
        """
        return BaseChecker.is_frozen(self, "OUTCAR")

    def is_complete(self):
        """Check if single VASP non-NEB calculation is complete.
        """
        usertime=False
        reachedaccuracy=False
        opath = os.path.join(self.keywords['name'],"OUTCAR")
        if not os.path.isfile(opath):
            self.logger.info("No OUTCAR at %s; not complete." % opath)
            return False

        myoutcar = Outcar(opath)
        
        #hw 04/19/13
        try:
            if myoutcar.run_stats['User time (sec)'] > 0:
                usertime=True
            else:
                usertime=False
        except KeyError:
            usertime=False
        
        reachgrep=subprocess.Popen('grep "reached required accuracy" %s' % opath, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        reachrpt=reachgrep.communicate()[0]
        reachgrep.wait()
        if reachrpt=='':
            reachedaccuracy=False
        else:
            reachedaccuracy=True

        isstatic=False
        isphonon=False
        isMD=False
        if 'ibrion' in self.keywords['program_keys'].keys():
            ibrionval = str(self.keywords['program_keys']['ibrion'])
            if ibrionval == "-1":
                self.logger.info("ingredient %s IBRION -1 is static" % self.keywords['name'])
                isstatic = True
            elif ibrionval == "0":
                self.logger.info("ingredient %s IBRION 0 is MD" % self.keywords['name'])
                isMD = True
            elif ibrionval in ["5","6","7","8"]:
                self.logger.info("ingredient %s IBRION %s is phonon" % (self.keywords['name'],ibrionval))
                isphonon = True
        if 'nsw' in self.keywords['program_keys'].keys():
            nswval = str(self.keywords['program_keys']['nsw'])
            if nswval in ["0","1"]:
                self.logger.info("ingredient %s NSW %s is static" % (self.keywords['name'],nswval))
                isstatic = True
        else:
            self.logger.info("ingredient %s no NSW given is static" % self.keywords['name'])
            isstatic = True #No NSW specified, VASP defaults to 0
        
        #For static runs, make an additional check for just electronic convergence.
        if isstatic:
            reachgrep=subprocess.Popen('grep "EDIFF is reached" %s' % opath, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            reachrpt=reachgrep.communicate()[0]
            reachgrep.wait()
            if reachrpt=='':
                reachedaccuracy=False
            else:
                reachedaccuracy=True


        if (isMD) or (isphonon):
            if not usertime:
                self.logger.warning("OUTCAR at %s shows no user time." % opath)
                return False
            else:
                self.logger.info("OUTCAR at %s shows user time." % opath)
                return True

        if isstatic:
            if reachedaccuracy:
                if usertime:
                    self.logger.info("OUTCAR at %s shows EDIFF reached for static run and user time; complete." % opath)
                    return True
                else:
                    self.logger.warning("OUTCAR at %s shows EDIFF reached for static run, but no user time." % opath)
                    return False
            else:
                if usertime:
                    self.logger.error("OUTCAR at %s does not show EDIFF reached for static run, but shows complete." % opath)
                    return False
                else:
                    self.logger.info("OUTCAR at %s does not show EDIFF reached for static run or user time; still incomplete." % opath)
                    return False

        else:
            if reachedaccuracy:
                if usertime:
                    self.logger.info("OUTCAR at %s shows reached required accuracy and user time; complete." % opath)
                    return True
                else:
                    self.logger.warning("OUTCAR at %s shows reached required accuracy, but no user time." % opath)
                    return False
            else:
                if usertime:
                    self.logger.error("OUTCAR at %s does not show reached required accuracy, but shows complete." % opath)
                    return False
                else:
                    self.logger.info("OUTCAR at %s does not show reached required accuracy or user time; still incomplete." % opath)
                    return False

    def is_ready_to_run(self):
        """Check if single VASP non-NEB ingredient is 
            ready to run.
        """
        dirname = self.keywords['name']
        notready=0
        if not(os.path.isfile(dirname + "/KPOINTS")):
            notready = notready + 1
        if not(os.path.isfile(dirname + "/POTCAR")):
            notready = notready + 1
        if not(os.path.isfile(dirname + "/INCAR")):
            notready = notready + 1
        if not(os.path.isfile(dirname + "/POSCAR")):
            notready = notready + 1
        if not(os.path.isfile(dirname + "/submit.sh")):
            notready = notready + 1
        if notready > 0:
            return False
        else:
            return True

    def _vasp_poscar_setup(self):
        """Set up the POSCAR file for a single VASP run.
        """
        name = self.keywords['name']
        pospath = os.path.join(name, "POSCAR")
        if os.path.isfile(pospath):
            my_poscar = Poscar.from_file(pospath) 
            #parent should have given a structure
        else: #this is an originating run; mast should give it a structure
            my_poscar = Poscar(self.keywords['structure'])
            self.logger.info("No POSCAR found from a parent; base structure used for %s" % self.keywords['name'])
        if 'mast_coordinates' in self.keywords['program_keys'].keys():
            sxtend = StructureExtensions(struc_work1=my_poscar.structure, name=self.keywords['name'])
            coordstrucs=self.get_coordinates_only_structure_from_input()
            newstruc = sxtend.graft_coordinates_onto_structure(coordstrucs[0])
            my_poscar.structure=newstruc.copy()
        dirutil.lock_directory(name)
        my_poscar.write_file(pospath)
        dirutil.unlock_directory(name)
        return my_poscar

    def _vasp_kpoints_setup(self):
        """Parse mast_kpoints string, which should take the format:
            number, number, number designation
            examples: "3x3x3 M", "1x1x1 G". If no designation is given,
            Monkhorst-Pack is assumed.
        """
        name = self.keywords['name']
        tryname = os.path.join(name, "KPOINTS")
        if os.path.isfile(tryname):
            #KPOINTS already exists. Do not overwrite.
            my_kpoints = Kpoints.from_file(tryname)
            return my_kpoints
        if not (self.metafile.read_data('kpoints')==None):
            kpoints = self.metafile.read_data('kpoints').split()
            kmesh = (int(kpoints[0].split('x')[0]),int(kpoints[0].split('x')[1]),int(kpoints[0].split('x')[2]))
            if len(kpoints) == 4:
                desig = "M"
                kshift = (float(kpoints[1]),float(kpoints[2]),float(kpoints[3]))
            else:
                desig = kpoints[1].upper()
                kshift = (float(kpoints[2]),float(kpoints[3]),float(kpoints[4]))
        else:
            if 'mast_kpoints' in self.keywords['program_keys'].keys():
                kpoints = self.keywords['program_keys']['mast_kpoints']
            else:
                raise MASTError(self.__class__.__name__,"k-point instructions need to be set either in ingredients keyword mast_kpoints or scaling section in structure ingredient: No k-point settings for the ingredient %s"% name)
            if len(kpoints) == 3:
                desig = "M"
            else:
                desig = kpoints[3].upper()
            kmesh = (int(kpoints[0]),int(kpoints[1]),int(kpoints[2]))
            kshift = (0,0,0)
        if desig == "M":
            my_kpoints = Kpoints.monkhorst_automatic(kpts=kmesh,shift=kshift)
        elif desig == "G":
            my_kpoints = Kpoints.gamma_automatic(kpts=kmesh,shift=kshift)
        else:
            raise MASTError(self.__class__.__name__,"kpoint designation " + desig + " not recognized.")

        dirutil.lock_directory(name)
        my_kpoints.write_file(name + "/KPOINTS")
        dirutil.unlock_directory(name)
        return my_kpoints

    def _vasp_potcar_setup(self, my_poscar):
        """Set up the POTCAR file."""
        name = self.keywords['name']

        if 'mast_xc' in self.keywords['program_keys'].keys():
            myxc = self.keywords['program_keys']['mast_xc'].upper() #Uppercase
        else:
            raise MASTError("vasp_checker, _vasp_potcar_setup","Exchange correlation functional needs to be specified in ingredients keyword mast_xc")

        if ('mast_pp_setup' in self.keywords['program_keys'].keys()):
            sites = my_poscar.site_symbols
            setup = self.keywords['program_keys']['mast_pp_setup']
            pp_sites = list()
            for site in sites:
                try:
                    pp_sites.append(setup[site])
                except KeyError:
                    pp_sites.append(site)
            my_potcar = Potcar(symbols=pp_sites, functional=myxc, sym_potcar_map=None)
        else:
            my_potcar = Potcar(symbols=my_poscar.site_symbols, functional=myxc, sym_potcar_map=None)

        dirutil.lock_directory(name)
        my_potcar.write_file(name + "/POTCAR")
        dirutil.unlock_directory(name)
        return my_potcar

    def _vasp_incar_get_non_mast_keywords(self):
        """Get the non-VASP keywords and make a dictionary."""
        incar_dict=dict()
        allowedpath = os.path.join(dirutil.get_mast_control_path(),
                        'programkeys','vasp_allowed_keywords.py')
        allowed_list = self._vasp_incar_get_allowed_keywords(allowedpath)
        for key, value in self.keywords['program_keys'].iteritems():
            if not key[0:5] == "mast_":
                keytry = key.upper()
                if not (keytry in allowed_list):
                    self.logger.warning("Ignoring program key %s for INCAR. To allow this keyword, add it to %s" % (keytry, allowedpath))
                else:
                    if type(value)==str and value.isalpha():
                        incar_dict[keytry]=value.capitalize() #First letter cap
                    else:
                        incar_dict[keytry]=value
        return incar_dict


    def _vasp_incar_get_allowed_keywords(self, allowedpath):
        """Get allowed vasp keywords.
            Args:
                allowedpath <str>: file path for allowed vasp keywords
        """
        allowed = MASTFile(allowedpath)
        allowed_list=list()
        for entry in allowed.data:
            allowed_list.append(entry.strip())
        return allowed_list

    def _vasp_incar_setup(self, my_potcar, my_poscar):
        """Set up the INCAR, including MAGMOM string, ENCUT, and NELECT."""
        name=self.keywords['name']
        myd = dict()
        myd = self._vasp_incar_get_non_mast_keywords()
        try:
            myd.pop("IMAGES")
        except KeyError:
            pass
        if 'mast_multiplyencut' in self.keywords['program_keys'].keys():
            mymult = float(self.keywords['program_keys']['mast_multiplyencut'])
        else:
            mymult = 1.5
        if 'ENCUT' in myd.keys():
            pass
        else:
            myd['ENCUT']=self._get_max_enmax_from_potcar(my_potcar)*mymult
        if 'mast_setmagmom' in self.keywords['program_keys'].keys():
            magstr = str(self.keywords['program_keys']['mast_setmagmom'])
            magmomstr=""
            maglist = magstr.split()
            numatoms = sum(my_poscar.natoms)
            if len(maglist) < numatoms:
                magct=0
                while magct < len(maglist):
                    magmomstr = magmomstr + str(my_poscar.natoms[magct]) + "*" + maglist[magct] + " " 
                    magct = magct + 1
            else:
                magmomstr = magstr
            myd['MAGMOM']=magmomstr
        if 'mast_charge' in self.keywords['program_keys'].keys():
            myelectrons = self.get_total_electrons(my_poscar, my_potcar)
            newelectrons=0.0
            try:
                adjustment = float(self.keywords['program_keys']['mast_charge'])
            except (ValueError, TypeError):
                raise MASTError("vasp_checker, vasp_incar_setup","Could not parse adjustment")
            #newelectrons = myelectrons + adjustment
            newelectrons = myelectrons - adjustment
            myd['NELECT']=str(newelectrons)
        my_incar = Incar(myd)
        dirutil.lock_directory(name)
        my_incar.write_file(name + "/INCAR")
        dirutil.unlock_directory(name)
        return my_incar

    def set_up_program_input(self):
        """Set up the program input files."""
        myposcar = self._vasp_poscar_setup()
        self._vasp_kpoints_setup()
        mypotcar = self._vasp_potcar_setup(myposcar)
        self._vasp_incar_setup(mypotcar, myposcar)
        return

    def forward_extra_restart_files(self, childpath):
        """Forward extra restart files: 
            For VASP, this entails a softlink to WAVECAR and 
            CHGCAR.
            Args:
                childpath <str>: path to child ingredient
        """
        raise MASTError(self.__class__.__name__,"This function is obsolete. Use softlink_charge_density_file, softlink_wavefunction_file, forward_charge_density_file, and/or forward_wavefunction_file")
        parentpath = self.keywords['name']
        dirutil.lock_directory(childpath)
        import subprocess
        #print "cwd: ", os.getcwd()
        curpath = os.getcwd()
        os.chdir(childpath)
        #print parentpath
        #print childpath
        mylink=subprocess.Popen("ln -s %s/WAVECAR WAVECAR" % parentpath, shell=True)
        mylink.wait()
        mylink2=subprocess.Popen("ln -s %s/CHGCAR CHGCAR" % parentpath, shell=True)
        mylink2.wait()
        os.chdir(curpath)
        dirutil.unlock_directory(childpath)

    def softlink_charge_density_file(self, childpath):
        """Softlink the parent charge density file to the child folder
            For VASP, this entails a softlink to CHGCAR
            Args:
                childpath <str>: path to child ingredient
        """
        return self.softlink_a_file(childpath, "CHGCAR")
    
    def softlink_wavefunction_file(self, childpath):
        """Softlink the parent wavefunction file to the child folder
            For VASP, this entails a softlink to WAVECAR.
            Args:
                childpath <str>: path to child ingredient
        """
        return self.softlink_a_file(childpath, "WAVECAR")

    def forward_charge_density_file(self, childpath):
        """Copy the charge density file to the child folder.
            For VASP, this is the CHGCAR.
            Args:
                childpath <str>: path to child ingredient
        """
        return self.copy_a_file(childpath, "CHGCAR","CHGCAR")
    def forward_wavefunction_file(self, childpath):
        """Copy the wavefunction file to the child folder.
            For VASP, this is the WAVECAR.
            CHGCAR.
            Args:
                childpath <str>: path to child ingredient
        """
        return self.copy_a_file(childpath, "WAVECAR","WAVECAR")
    
    def add_selective_dynamics_to_structure_file(self, sdarray):
        """Add selective dynamics to a structure file.
            In the case of VASP, the structure file is 
            POSCAR-like
            Args:
                sdarray <numpy array of bool>: Numpy array of
                    booleans for T F F etc.
        """
        name = self.keywords['name']
        pname = os.path.join(name,"POSCAR")
        phposcar = Poscar.from_file(pname)
        phposcar.selective_dynamics = sdarray
        dirutil.lock_directory(name)
        os.rename(pname, pname + "_no_sd")
        phposcar.write_file(pname)
        dirutil.unlock_directory(name)
        return

    def get_energy(self):
        """Get the energy.
            For VASP, this is E0 energy from vasprun.xml
        """
        return Vasprun('%s/vasprun.xml' % self.keywords['name']).ionic_steps[-1]["electronic_steps"][-1]["e_0_energy"]

    def _get_max_enmax_from_potcar(self, mypotcar):
        """Get maximum enmax value (float) from Potcar 
            (combined list)
        """
        enmax_list=list()
        potcarct=0
        onepotcar=None
        while potcarct < len(mypotcar):
            onepotcar = mypotcar[potcarct] #A PotcarSingle object
            enmax_list.append(onepotcar.enmax)
            potcarct = potcarct + 1
        return max(enmax_list)

    def _make_one_unfrozen_atom_poscar(self, myposcar, natom):
        """Use selective dynamics to make a poscar with one unfrozen atom.
            myposcar = Poscar
            natom = the number of the atom to unfreeze
            Returns: Poscar (use write_file function on it).
        """
        raise MASTError(self.__class__.__name__, "This method is abandoned and should not be used.")
        mysd=np.zeros([sum(myposcar.natoms),3],bool)
        mysd[natom-1][0]=True #indexing starts at 0
        mysd[natom-1][1]=True
        mysd[natom-1][2]=True
        myposcar.selective_dynamics = mysd
        return myposcar

    def _make_one_unfrozen_direction_poscar(self, myposcar, natom, ndir):
        """Use selective dynamics to make a poscar with one unfrozen atom.
            myposcar = Poscar
            natom = the number of the atom to unfreeze
            ndir = the direction to freeze (0, 1, 2 for x, y, z)
            Returns: Poscar (use write_file function on it).
        """
        raise MASTError(self.__class__.__name__, "This method is abandoned and should not be used.")
        mysd=np.zeros([sum(myposcar.natoms),3],bool)
        mysd[natom-1][ndir]=True #indexing starts at 0
        myposcar.selective_dynamics = mysd
        return myposcar


    def read_my_dynamical_matrix_file(self, mydir="", fname="DYNMAT"):
        """Read a dynamical matrix file.
            For VASP this is DYNMAT.
            Args:
                mydir <str>: directory; use ingredient directory
                             if null is given
                fname <str>: file name (default 'DYNMAT')
            Returns:
                dyndict <dict>: dictionary structured like this:
                    ['numspec'] = <int> number of species
                    ['numatoms'] = <int> number of atoms
                    ['numdisp'] = <int> number of displacements
                    ['massline'] = <str> masses line
                    ['atoms'][atom <int>][disp <int>]['displine'] =
                        displacement
                        vector (part of first line in dynmat block,
                        e.g. "0.01 0 0")
                    ['atoms'][atom <int>][disp <int>]['dynmat'] =
                            <list>
                            list of dynmat lines for this atom
                            and this displacement
        """
        if mydir == "":
            mydir = self.keywords['name']
        mydyn = MASTFile(os.path.join(mydir, fname))
        mydata = list(mydyn.data) #whole new copy
        dyndict=dict()
        firstline = mydata.pop(0) #pop first value
        firstspl = firstline.strip().split()
        dyndict['numspec']=int(firstspl[0])
        numatoms = int(firstspl[1])
        dyndict['numatoms']=numatoms
        dyndict['numdisp']=int(firstspl[2])
        dyndict['massline']=mydata.pop(0)
        dyndict['atoms']=dict()
        atom=0
        disp=0
        thirdline=""
        while len(mydata) > 0:
            thirdline = mydata.pop(0)
            thirdspl = thirdline.strip().split()
            atom=int(thirdspl[0])
            disp=int(thirdspl[1])
            displine=' '.join(thirdspl[2:])
            if not atom in dyndict['atoms'].keys():
                dyndict['atoms'][atom]=dict()
            dyndict['atoms'][atom][disp]=dict()
            dyndict['atoms'][atom][disp]['displine'] = displine
            dyndict['atoms'][atom][disp]['dynmat']=list()
            for act in range(0, numatoms):
                dyndict['atoms'][atom][disp]['dynmat'].append(mydata.pop(0))
        return dyndict

    def write_my_dynamical_matrix_file(self, dyndict, mydir="", fname="DYNMAT"):
        """Write a dynamical matrix file based on a dictionary.
            Args:
                dyndict <dict>: Dictionary of dynmat (see read_my_dynmat)
                mydir <str>: Directory in which to write;
                             use ingredient directory if null
                fname <str>: filename (default DYNMAT)
        """
        if mydir == "":
            mydir = self.keywords['name']
        dynwrite=MASTFile()
        dynwrite.data=list()
        firstline=str(dyndict['numspec']) + " " + str(dyndict['numatoms']) + " " + str(dyndict['numdisp']) + "\n"
        dynwrite.data.append(firstline)
        dynwrite.data.append(dyndict['massline'])
        atomlist=dyndict['atoms'].keys()
        atomlist.sort()
        for atom in atomlist:
            displist = dyndict['atoms'][atom].keys()
            displist.sort()
            for disp in displist:
                thirdline = str(atom) + " " + str(disp) + " " + dyndict['atoms'][atom][disp]['displine'] + "\n"
                dynwrite.data.append(thirdline)
                for line in dyndict['atoms'][atom][disp]['dynmat']:
                    dynwrite.data.append(line)
        dynwrite.to_file(os.path.join(mydir, fname))


    def write_my_dynmat_without_disp_or_mass(self, dyndict, mydir="", fname="DYNMAT"):
        """Write a dynamical matrix file without the displacement indicators 1, 2, 3
            and without the masses line, and with first line having only
            the total number of displacements, for PHON.
            Args:
                dyndict <dict>: Dictionary of dynmat (see read_my_dynmat)
                mydir <str>: Directory in which to write; use 
                             ingredient directory if null
                fname <str>: filename (default DYNMAT)
        """
        if mydir == "":
            mydir = self.keywords['name']
        dynwrite=MASTFile()
        dynwrite.data=list()
        firstline=str(dyndict['numdisp']) + "\n"
        dynwrite.data.append(firstline)
        atomlist=dyndict['atoms'].keys()
        atomlist.sort()
        for atom in atomlist:
            displist = dyndict['atoms'][atom].keys()
            displist.sort()
            for disp in displist:
                thirdline = str(atom) + " " + dyndict['atoms'][atom][disp]['displine'] + "\n"
                dynwrite.data.append(thirdline)
                for line in dyndict['atoms'][atom][disp]['dynmat']:
                    dynwrite.data.append(line)
        dynwrite.to_file(os.path.join(mydir, fname))

    def read_my_displacement_file(self, mydir="", fname="XDATCAR"):
        """Read a displacement file. For VASP this is XDATCAR.
            Args:
                mydir <str>: Directory. Use ingredient directory
                             if null.
                fname <str>: Filename to read. Default is
                             XDATCAR
            Returns:
                xdatdict <dict>: Dictionary of configurations
                    xdatdict['descline'] = <str> description line
                    xdatdict['specline'] = <str> species line
                    xdatdict['numline'] = <str> numbers lines
                    xdatdict['type'] = <str> Direct or not
                    xdatdict['numatoms'] = <int> number of atoms
                    xdatdict['configs'][config <int>] = <list>
                            list of lines in this configuration
                    xdatdict['scale'] = <str> scale
                    xdatdict['latta'] = <str> lattice vector a
                    xdatdict['lattb'] = <str> lattice vector b
                    xdatdict['lattc'] = <str> lattice vector c
        """
        if mydir == "":
            mydir = self.keywords['name']
        myxdat = MASTFile(os.path.join(mydir, fname))
        mydata = list(myxdat.data) #whole new copy
        xdatdict=dict()
        xdatdict['descline'] = mydata.pop(0) #pop first value
        tryspec = mydata.pop(0)
        tryspecstrip = tryspec.strip()
        tryspecstrip = tryspecstrip.replace(" ","x")
        if not tryspecstrip.isalpha(): # version 5.2.11 and on
            xdatdict['scale'] = tryspec
            xdatdict['latta'] = mydata.pop(0)
            xdatdict['lattb'] = mydata.pop(0)
            xdatdict['lattc'] = mydata.pop(0)
            xdatdict['specline'] = mydata.pop(0)
        else:
            xdatdict['scale'] = ""
            xdatdict['latta'] = ""
            xdatdict['lattb'] = ""
            xdatdict['lattc'] = ""
            xdatdict['specline'] = tryspec
        numline = mydata.pop(0)
        xdatdict['numline'] = numline
        numatoms = sum(map(int, numline.strip().split()))
        xdatdict['numatoms'] = numatoms
        if not tryspecstrip.isalpha(): #version 5.2.11 and on
            xdatdict['type'] = ""
        else:
            xdatdict['type'] = mydata.pop(0)
        xdatdict['configs']=dict()
        kfgct=1
        while len(mydata) > 0:
            mydata.pop(0) # Konfig line, or a blank line
            xdatdict['configs'][kfgct] = list()
            for act in range(0,numatoms):
                if len(mydata) > 0:
                    xdatdict['configs'][kfgct].append(mydata.pop(0))
            kfgct=kfgct+1
        return xdatdict

    def write_my_displacement_file(self, xdatdict, mydir="", fname="XDATCAR"):
        """Write a displacement file.
            Args:
                xdatdict <dict>: Dictionary of XDATCAR (see read_my_xdatcar)
                mydir <str>: Directory in which to write.
                             Use ingredient directory if null.
                fname <str>: filename (default DYNMAT)
        """
        if mydir == "":
            mydir = self.keywords['name']
        xdatwrite=MASTFile()
        xdatwrite.data=list()
        xdatwrite.data.append(xdatdict['descline'])
        if not (xdatdict['scale'] == ""):
            xdatwrite.data.append(xdatdict['scale'])
            xdatwrite.data.append(xdatdict['latta'])
            xdatwrite.data.append(xdatdict['lattb'])
            xdatwrite.data.append(xdatdict['lattc'])
        xdatwrite.data.append(xdatdict['specline'])
        xdatwrite.data.append(xdatdict['numline'])
        if not (xdatdict['type'] == ""):
            xdatwrite.data.append(xdatdict['type'])
        configlist = xdatdict['configs'].keys()
        configlist.sort()
        for cfg in configlist:
            xdatwrite.data.append("Konfig=%1i\n" % cfg)
            xdatwrite.data.extend(xdatdict['configs'][cfg])
        xdatwrite.to_file(os.path.join(mydir, fname))

    def combine_dynamical_matrix_files(self, mydir=""):
        """Combine dynamical matrix files into one file.
            The subfiles should be in subfolders under
            the top directory.
            Args:
                mydir <str>: top directory for DYNMAT files.
                             Use ingredient directory if null.
            Returns:
                Creates a single DYNMAT file
        """
        if mydir == "":
            mydir = self.keywords['name']
        dynmatlist = dirutil.walkfiles(mydir, 2, 5, "*DYNMAT*") #start one level below
        if len(dynmatlist) == 0:
            raise MASTError("pmgextend combine_dynmats", "No DYNMATs found under " + mydir)
        totnumdisp=0
        largedyn=dict()
        largedyn['atoms'] = dict()
        for onedynmat in dynmatlist:
            dyndir = os.path.dirname(onedynmat)
            onedyn = self.read_my_dynamical_matrix_file(dyndir)
            totnumdisp = totnumdisp + onedyn['numdisp']
            for atom in onedyn['atoms'].keys():
                if not atom in largedyn['atoms'].keys():
                    largedyn['atoms'][atom]=dict()
                    mydisp=1 #start at 1
                for disp in onedyn['atoms'][atom].keys():
                    if disp in largedyn['atoms'][atom].keys():
                        mydisp=mydisp + 1 #increment
                    largedyn['atoms'][atom][mydisp]=dict()
                    largedyn['atoms'][atom][mydisp]['displine'] = str(onedyn['atoms'][atom][disp]['displine'])
                    largedyn['atoms'][atom][mydisp]['dynmat']=list(onedyn['atoms'][atom][disp]['dynmat'])
        largedyn['numspec'] = onedyn['numspec'] #should all be the same
        largedyn['numatoms'] = onedyn['numatoms']
        largedyn['massline'] = onedyn['massline']
        largedyn['numdisp'] = totnumdisp
        self.write_my_dynamical_matrix_file(largedyn, mydir, "DYNMAT_combined")
        self.write_my_dynamical_matrix_file(largedyn, mydir, "DYNMAT")

    def combine_displacement_files(self, mydir=""):
        """Combine displacement files (here XDATCARs) into one file.
            Args:
                mydir <str>: top directory for DYNMAT files.
                             Use ingredient directory if null.
        """
        if mydir == "":
            mydir = self.keywords['name']
        largexdat=dict()
        xdatlist = dirutil.walkfiles(mydir, 2, 5, "*XDATCAR*") #start one level below
        if len(xdatlist) == 0:
            raise MASTError("pmgextend combine_displacements", "No XDATCARs found under " + mydir)
        kfgct=1 # skip config 1 until the end
        largexdat['configs']=dict()
        xdatlist.sort() #get them in order
        for onexdatmat in xdatlist:
            xdatdir = os.path.dirname(onexdatmat)
            onexdat = self.read_my_displacement_file(xdatdir)
            for kfg in onexdat['configs'].keys():
                if kfg == 1: #skip config 1 until the end
                    pass
                else:
                    kfgct = kfgct + 1 #start at 2
                    largexdat['configs'][kfgct] = onexdat['configs'][kfg]
        largexdat['configs'][1] = onexdat['configs'][1] #get one of the first configs (the first should be the same for all the XDATCARs)
        largexdat['descline'] = onexdat['descline'] #should all be the same
        largexdat['specline'] = onexdat['specline']
        largexdat['scale'] = onexdat['scale']
        largexdat['latta'] = onexdat['latta']
        largexdat['lattb'] = onexdat['lattb']
        largexdat['lattc'] = onexdat['lattc']
        largexdat['numline'] = onexdat['numline']
        largexdat['numatoms'] = onexdat['numatoms']
        largexdat['type'] = onexdat['type']
        self.write_my_displacement_file(largexdat, mydir, "XDATCAR_combined")
        self.write_my_displacement_file(largexdat, mydir, "XDATCAR")
    def make_hessian(self, myposcar, mydir):
        """Combine DYNMATs into one hessian and solve for frequencies.
            myposcar = Poscar
            mydir = top directory for DYNMAT files
        """
        raise MASTError(self.__class__.__name__, "This method is abandoned and should not be used.")
        natoms = sum(myposcar.natoms)
        myhess=np.zeros([natoms*3, natoms*3])
        #arrange as x1, y1, z1, x2, y2, z2, etc.
        dynmatlist = walkfiles(mydir, 1, 5, "*DYNMAT*")
        if len(dynmatlist) == 0:
            raise MASTError("pmgextend combine_dynmats", "No DYNMATs found under " + mydir)
        opendyn=""
        print "DYNMATLIST:"
        print dynmatlist
        datoms=0
        for onedynmat in dynmatlist:
            dynlines=[]
            opendyn = open(onedynmat,'rb')
            dynlines=opendyn.readlines()
            opendyn.close()
            datoms = int(dynlines[0].split()[1])
            dmats = int(dynlines[0].split()[2])
            mycount=2 #starting line
            while mycount < len(dynlines)-datoms:
                littlemat=[]
                topatom = int(dynlines[mycount].split()[0])
                whichdir = int(dynlines[mycount].split()[1])
                littlemat = dynlines[mycount+1:mycount+datoms+1]
                print littlemat
                act = 0
                while act < datoms:
                    dactx=float(littlemat[act].split()[0])
                    dacty=float(littlemat[act].split()[1])
                    dactz=float(littlemat[act].split()[2])
                    colidx = (topatom-1)*3 + (whichdir-1)
                    #so 2  3  on first line means atom 2's z direction
                    #then with atom 1x, 1y, 1z; 2x, 2y, 2z, etc.
                    myhess[colidx][act*3+0]=dactx
                    myhess[colidx][act*3+1]=dacty
                    myhess[colidx][act*3+2]=dactz
                    act = act + 1
                mycount = mycount + datoms + 1
                print mycount
        print "UNALTERED HESSIAN:"
        print(myhess)
        #create mass matrix
        masses=dynlines[1].split()
        print "MASSES:", masses
        massarr=np.zeros([datoms*3,1])
        act=0
        print myposcar.natoms
        nspec=len(myposcar.natoms)
        totatoms=0
        while act < datoms:
            mymass=0
            nct=0
            totatoms=0
            while (mymass==0) and nct < nspec:
                totatoms = totatoms + myposcar.natoms[nct]
                if act < totatoms:
                    mymass = float(masses[nct])
                nct = nct + 1
            print mymass
            massarr[act*3+0][0]=mymass
            massarr[act*3+1][0]=mymass
            massarr[act*3+2][0]=mymass
            act = act + 1
        massmat = massarr*np.transpose(massarr)
        print "MASS MAT:"
        print massmat
        print "STEP:"
        step = float(dynlines[2].split()[2])
        print step
        print "HESSIAN * -1 / step / sqrt(mass1*mass2)"
        normhess=np.zeros([natoms*3,natoms*3])
        cidx=0
        while cidx < natoms*3:
            ridx=0
            while ridx < natoms*3:
                normhess[ridx][cidx]=-1*myhess[ridx][cidx]/step/np.sqrt(massmat[ridx][cidx])
                ridx = ridx + 1
            cidx = cidx + 1
        print normhess
        print "EIGENVALUES:"
        myeig = np.linalg.eig(normhess)[0]
        print myeig
        print "SQRT of EIGENVALUES in sqrt(eV/AMU)/Angstrom/2pi:"
        myfreq = np.sqrt(myeig)
        print myfreq
        print "SQRT OF EIGENVALUES in THz:"
        myfreqThz = myfreq*15.633302
        print myfreqThz
        myfreqThzsorted = myfreqThz
        myfreqThzsorted.sort()
        print myfreqThzsorted
        return myfreqThzsorted

    def get_total_electrons(self, myposcar, mypotcar):
        """Get the total number of considered electrons in the system."""
        atomlist = myposcar.natoms
        zvallist = self.get_valence_list(mypotcar)
        totzval = 0.0
        atomct = 0
        if not (len(zvallist) == len(atomlist)):
            raise MASTError(self.__class__.__name__,
                "Number of species and number of POTCARs do not match.")
        while atomct < len(atomlist):
            totzval = totzval + (atomlist[atomct] * zvallist[atomct])
            atomct = atomct + 1
        return totzval

    def get_valence_list(self, mypotcar):
        """Get list of number of valence electrons for species.
            For VASP, this is a list of zvals from POTCAR.
            Args:
                mypotcar <list of PotcarSingle objects>
            Returns:
                zval_list <list of int>: List of number of
                                        valence electrons
        """
        zval_list=list()
        potcarct=0
        onepotcar=None
        while potcarct < len(mypotcar):
            onepotcar = mypotcar[potcarct] #A PotcarSingle object
            zval_list.append(onepotcar.zval)
            potcarct = potcarct + 1
        return zval_list
    def get_energy_from_energy_file(self):
        """Get the energy from the energy file.
            For VASP, this is the last E0 energy from OSZICAR,
            and this function should be used if vasprun.xml
            is corrupted or not available.
            Args:
                mydir <str>: Directory in which to look.
            Returns:
                <float>: last E0 energy from OSZICAR
        """
        fullpath=os.path.join(self.keywords['name'], "OSZICAR")
        if not os.path.isfile(fullpath):
            raise MASTError(self.__class__.__name__, "No OSZICAR file at %s" % self.keywords['name'])
        myosz = MASTFile(fullpath)
        mye0 = myosz.get_segment_from_last_line_match("E0", "E0=","d E =")
        mye0float=""
        try:
            mye0float=float(mye0)
        except TypeError:
            self.logger.error("Failed to log energy %s" % str(mye0))
        return mye0float
    def is_started(self):
        """See if the ingredient has been started on
            the queue.
        """
        if os.path.isfile(os.path.join(self.keywords['name'],'OUTCAR')):
            return True
        else:
            return False

    def write_final_structure_file(self, mystruc):
        """Write the final structure to a file.
            For VASP, this is CONTCAR.
        """
        mycontcar = Poscar(mystruc)
        mycontcar.write_file(os.path.join(self.keywords['name'],'CONTCAR'))
    def has_starting_structure_file(self):
        """Evaluate whether the ingredient has a starting
            structure file. For VASP, this is a POSCAR.
        """
        return os.path.isfile(os.path.join(self.keywords['name'], 'POSCAR'))

    def has_ending_structure_file(self):
        """Evaluate whether the ingredient has a starting
            structure file. For VASP, this is a CONTCAR.
        """
        return os.path.isfile(os.path.join(self.keywords['name'], 'CONTCAR'))
    def scale_mesh(self, newstr, density):
        """Scale the kpoint mesh.
            Args:
                newstr <structure>: new structure
                density <float>: kpoint density
            Returns:
                newkmesh <pymatgen Kpoints>: new kpoint mesh
                sets program_keys 'mast_kpoints' to new mesh
        """
        newkmesh = Kpoints.automatic_density(newstr, density)
        klist = list()
        klist = newkmesh.to_dict['kpoints'][0]
        klist.append(newkmesh.to_dict['generation_style'][0])
        self.keywords['program_keys']['mast_kpoints'] = klist
        return newkmesh

    def get_final_pressure(self):
        """Get the final pressure.
            For VASP, this is the last pressure line from
            the OUTCAR.
            Args:
                mydir <str>: Directory in which to look.
            Returns:
                <float>: last pressure from OUTCAR, in kB
        """
        fullpath=os.path.join(self.keywords['name'], "OUTCAR")
        if not os.path.isfile(fullpath):
            raise MASTError(self.__class__.__name__, "No OUTCAR file at %s" % self.keywords['name'])
        myoutcar = MASTFile(fullpath)
        mypress = myoutcar.get_segment_from_last_line_match("pressure", "external pressure =","kB  Pullay stress =")
        mypressfloat=""
        try:
            mypressfloat=float(mypress)
        except TypeError:
            self.logger.error("Failed to log pressure %s" % str(mypress))
        return mypressfloat



import random
import logging

def moves_switch(indiv, Optimizer):
    """Select mutation to preform on individual from list of options"""
    #logger = initialize_logger(Optimizer.loggername)
    logger = logging.getLogger(Optimizer.loggername)
    scheme = random.choice(Optimizer.mutation_options)
    if 'MU' in Optimizer.debug:
        debug = True
    else:
        debug = False
    try:
       exec "from MAST.structopt.moves.{0} import {0}".format(scheme)
       mutant = eval('{0}(indiv, Optimizer)'.format(scheme))
       mutant.energy = 0
       mutant.fitness = 0
    except NameError, e:
        logger.warning('Specified mutation not one of the available options. Please check documentation and spelling! SKIPPING. Mutation : {0}. {1}'.format(scheme,e), exc_info=True)
        print 'Mutation Name error: ',e
        print scheme
        mutant = indiv
    except Exception, e:
        logger.error('Problem with mutation! SKIPPING. Mutation = {0}. {1}'.format(scheme,e), exc_info=True)
        print 'Mutation Exception: ', e
        mutant = indiv
    return mutant, scheme



##############################################################
# This code is part of the MAterials Simulation Toolkit (MAST)
# 
# Maintainer: Tam Mayeshiba
# Last updated: 2014-04-25
##############################################################
import os
import time
from MAST.utility import dirutil
from MAST.utility import MASTError

class MASTFile:
    """Controls MAST file IO and manipulation"""
    #data = [] #TTM 11/16/11 remove from here to avoid confusing reinit.

    #TTM 10/31/11 allow to initialize without a file path
    def __init__(self, file_path=str()):
        self.data = list() # TTM 11/16/11 add here
        self.p = dict() # TTM 12/22/11 set in file class instead #dictionary holding a mapping of param names to param values
        if file_path != "":
            if not os.path.isfile(file_path):
                raise MASTError(self.__class__.__name__,
                "No file at %s; cannot create MASTFile" % file_path)
            self.from_file(file_path)
        else:
            pass

    def from_file(self, file_path):
        """Reads data from a file and stores it in a list"""
        self.data = []
        #TTM+2 10/7/11 add error checking in case of no file
        if not os.path.isfile(file_path):
            raise MASTError(self.__class__.__name__,
                "No such file at " + file_path)
        readf = open(file_path,'rb')
        lines = readf.readlines()
        for line in lines:
            self.data.append(line)
        readf.close()    
    
    #TTM 12/6/11 create for multiple species potcars
    def from_file_append(self, file_path=str()):
        """Reads data from a file and appends it to the list"""
        if not os.path.isfile(file_path):
            return
        readf = open(file_path,'rb')
        lines = readf.readlines()
        for line in lines:
            self.data.append(line)
        readf.close()    
    
    def to_file(self,file_path):
        """Writes data to a file (overwrites existing file)."""
        #TTM+2 10/7/11 add error checking in case of no path
        if (file_path == "") or (file_path == None):
            raise MASTError(self.__class__.__name__,
                "Refusing to copy to empty path.")
        if self.data == []:
            raise MASTError(self.__class__.__name__,
                "Empty file not copied to " + file_path)
        dirutil.lock_directory(os.path.dirname(file_path))
        writef = open(file_path,'wb')
        for line in self.data:
            writef.write(line)
        writef.close()
        dirutil.unlock_directory(os.path.dirname(file_path))

    #TTM 11/10/11 created to solve archiving problems
    def to_unique_file(self, parent_path="", try_name="", suffix="", max=10):
        """Writes to a unique file (?)"""
        if (parent_path == "") or (parent_path == None):
            return
        if (try_name == "") or (try_name == None):
            try_name = str(time.time())
        basicname = os.path.join(parent_path, try_name)
        tryme = basicname + suffix
        ct = 1
        while os.path.isfile(tryme):
            tryme = basicname + str(ct) + suffix
            ct = ct + 1
            if ct > max: #TTM 11/23/11 increase max threshold
                print 'Over', max, 'of the same file. Erroring out.'
                return
        self.to_file(tryme)
        return tryme #TTM 20121109 return the file name

    def to_stdout(self):
        """Writes to stdout"""
        for line in self.data:
            print line

    #TTM+block 10/5/11 created (Pgetline)
    def get_line_number(self, line_number):
        """Grabs what line is at number line_number"""
        matidx = line_number - 1 # matrix indices start at 0
        if matidx < 0:
            return None
        if len(self.data) > matidx:
            return self.data[line_number - 1]
        else:
            return None

    #TTM+block 10/5/11 created (Pgetline)
    def get_line_match(self, string_to_match):
        """Searches for a string, and returns it once found"""
        for line in self.data:
            if string_to_match in line:
                return line
        return None

    #TTM+block 11/7/11 created to return line number
    def get_line_match_number(self, string_to_match):
        linect = len(self.data)
        ct = 0
        myline=""
        while ct < linect:
            myline = self.data[ct]
            if string_to_match in myline:
                return ct + 1 #return line numbers starting from 1
            ct = ct + 1

    #TTM+block 10/5/11 created
    def get_last_line_match(self, string_to_match):
        lastmatch=""
        for line in self.data:
            if string_to_match in line:
                lastmatch = line
        if lastmatch == "":
            return None
        return lastmatch

    def get_last_x_lines_line_match(self, string_to_match, x_lines):
        lastmatch=""
        x_int = int(x_lines)
        if x_int == 0:
            return self.get_last_line_match(string_to_match)
        searchme = self.data[-1*x_int:]
        for line in searchme:
            if string_to_match in line:
                lastmatch = line
        if lastmatch == "":
            return None
        return lastmatch

    #TTM+block 11/8/11 separated from get_segment_from_last_line_match
    def get_segment(self, string_to_chop, start_string="", end_string=""):
        start_found=""
        startpos=""
        endpos=""
        if start_string == "":
            startpos = 0 # start from beginning
        else:
            start_found = string_to_chop.find(start_string)
            if start_found == -1:
                return None # not found
            startpos = start_found + len(start_string) # start after the string
        if end_string == "":
            endpos = len(string_to_chop) # go to end
        else:
            endpos = string_to_chop.find(end_string)
            if endpos == -1:
                return None #not found
        return string_to_chop[startpos:endpos].strip() # strip off white space


    #TTM+block 10/5/11 created (Pgetseg)
    def get_segment_from_last_line_match(self, string_to_match, start_string="", end_string="", match_last_line=True):
        found=""
        if match_last_line:
            found = self.get_last_line_match(string_to_match)
        else:
            found = self.get_line_match(string_to_match)
        if not found:
            return None
        myseg = self.get_segment(found, start_string, end_string)
        return myseg

    #TTM 11/7/11 from Pmodfilenew.py
    def modify_file_by_line_number(self, lineno="", mode="", param=""):
        """Modifies a file at a line number"""
        #mode: D - Delete, R - Replace, I - insert
        #param: if mode is:
        #       D - ignored; leave blank
        #       R - string with which to replace whole line at lineno
        #       I - string to insert below lineno as a whole line
        try:
            lineno = int(lineno)
        except (ValueError, TypeError):
            print 'Invalid line number.'
            raise RuntimeError('Line number does not exist')

        try:
            mode = str.upper(mode)
        except (ValueError, TypeError):
            print 'Invalid mode.'
            raise RuntimeError('Selected mode does not exist')

        if mode not in ['D','R','I']:
            print 'Only D-delete, R-replace, I-insert supported.'
            return

        dataidx = lineno - 1 #sync line number to matrix elements
        origlen = len(self.data)
        before=[]
        after=[]

        if (dataidx > origlen - 1):
            print 'File does not contain', lineno, 'lines.'
            return

        if (dataidx < 0):
            print 'Negative line number cannot be found.'
            return
        #
        if (dataidx == 0):
            before=[]
        else:
            before = self.data[0:dataidx]

        if (dataidx == origlen - 1):
            after=[]
        else:
            after = self.data[dataidx+1:]

        if (mode == 'D'):
            pass # no special action
        elif (mode == 'R'):
            before.append(param)# put new line where old one went
        elif (mode == 'I'):
            before.append(self.data[dataidx]) # put line back in
            before.append(param)

        self.data = before + after
        return
    
    #TTM+block 11/8/11 created
    #TTM 12/22/11 moving to file class from INCAR class
    def file_to_dictionary(self, from_path):
        self.from_file(from_path)
        line = ""
        linesplit = list()
        for line in self.data:
            if line == "":
                pass #empty space
            if not ('=' in line):
                pass # system comment
            if line[0] == "#":
                pass #comments
            linesplit = line.split('=',1) # split in 1 equals sign
            if len(linesplit) > 1:
                self.p[linesplit[0].strip()] = linesplit[1].strip()

    #TTM 052112 added
    def copy_data_to(self, other_file):
        other_file.data = list()
        for line in self.data:
            other_file.data.append(line)
        return



import sys
import os
import time
import random
import math
import pdb
import logging
from MAST.structopt import inp_out
from MAST.structopt import tools
from MAST.structopt import generate
from MAST.structopt import switches
from MAST.structopt.switches.predator_switch import predator_switch
from MAST.structopt import fingerprinting
from MAST.structopt import post_processing as pp
from ase import Atom, Atoms
try:
    from mpi4py import MPI
except ImportError:
    pass

class Optimizer():
    __version__  = 'StructOpt_v0.4'
    logger = None
    
    def __init__(self, input, uselogger=True):
        if input:
            parameters = inp_out.read_parameter_input(input, uselogger)
        else:
            parameters = inp_out.read_parameter_input({'atomlist':[('Xx',1,0,0)],'structure':'Cluster'}, uselogger)
        self.__dict__.update(parameters)
        if self.loggername:
            global logger
            logger = logging.getLogger(self.loggername)
        if self.restart_optimizer:
            try:
                rank = MPI.COMM_WORLD.Get_rank()
            except:
                rank = 0
            if rank==0:
                logger.info('restarting output')
                outdict = inp_out.restart_output(self)
                self.__dict__.update(outdict)
                logger.info('Loading individual files')
                poplist = []
                for indfile in self.population:
                    ind = inp_out.read_individual(indfile)
                    poplist.append(ind)
                self.population = poplist
                logger.info('Loading bests')
                bestlist = []
                for bestfile in self.BESTS:
                    ind = inp_out.read_individual(bestfile)
                    bestlist.append(ind)
                self.BESTS = bestlist
                self.restart = True
                if self.structure == 'Defect':
                    bulk = inp_out.read_xyz(self.solidfile)
                    bulk.set_pbc(True)
                    bulk.set_cell(self.solidcell)
                    self.solidbulk = bulk.copy()
                else:
                    self.solidbulk = None
        else:
            self.convergence = False
            self.generation = 0
            self.Runtimes = [time.time()]
            self.Evaluations = list()
            self.CXs = list()
            self.Muts = list()
            self.cxattempts = 0
            self.mutattempts = list()
            self.BESTS = list()
            self.genrep = 0
            self.minfit = 0
            self.convergence = False
            self.overrideconvergence = False
            self.population = list()
            self.calc = None
            self.static_calc = None
    
    def algorithm_initialize(self):
        global logger
        if self.restart_optimizer:
            logger.info('Successfully loaded optimizer from file')
        else:
            #Setup the output files
            logger.info('Initializing output for algorithm')
            outdict = inp_out.setup_output(self.filename, self.restart, self.nindiv,
                self.indiv_defect_write, self.genealogy, self.allenergyfile,
                self.fingerprinting, self.debug)
            self.__dict__.update(outdict)
            #Set starting convergence and generation
            logger.info('Initializing convergence, generation, and output monitoring stats')
            self.convergence = False
            self.overrideconvergence = False
            self.generation = 0
            #Prep output monitoring
            self.Runtimes = [time.time()]
            self.Evaluations = list()
            self.CXs = list()
            self.Muts = list()
            self.cxattempts = 0
            self.mutattempts = list()
            #Initialize random number generator
            random.seed(self.seed)
            #Initialize swap list variable
            if self.swaplist == None:
                #Use if concentrations in swaplist is unknown
                self.swaplist = [ [sym,c] for sym,c,m,u in self.atomlist ]
                swaplist = self.swaplist
            elif self.swaplist:
                swaplist = self.swaplist
            #Write the input parameters to the output file
            logger.info('Writing the input parameters to output file')
            inp_out.write_parameters(self)
    
    def algorithm_island(self):
        """Subprogram to run the GA in a Island Model Parallel Scheme"""
        global logger
        comm = MPI.COMM_WORLD
        rank = MPI.COMM_WORLD.Get_rank()
        if 'MA' in self.debug:
            debug = True
        else:
            debug = False
        if rank==0:
            #Initialize Random - Ensure each population has different random seed
            seeds=[]
            while len(seeds) < comm.Get_size():
                a=random.randint(0,100)
                if a not in seeds:
                    seeds.append(a)
        else:
            seeds=None
        self = comm.bcast(self,root=0)
        seed = comm.scatter(seeds,root=0)
        self.seed=seed
        random.seed(self.seed)
        self.algorithm_initialize()
        logger.info('Beginning main algorithm loop')
        logger.info('Random seed = {0}'.format(seed))
        MIGEVENT=0
        while True:
            pop = self.population
            #Migration
            if self.generation != 0:
                if float(self.generation)/float(self.migration_intervals)%1==0:
                    orcs=comm.allgather(self.overrideconvergence)
                    if True in orcs:
                        break
                    logger.info('Migrating structures')
                    self.output.write('-----Migration in generation '+repr(self.generation)+'-----\n')
                    nmig=int(self.nindiv*self.migration_percent)
                    migrantsind=[]
                    for i in range(nmig):
                        migrantsind.append(pop[i].duplicate())
                    self.output.write('Migrating '+repr(len(migrantsind))+' individuals\n')
                    migrantsind=[migrantsind]
                    migrants = comm.gather(migrantsind,root=0)
                    if rank==0:
                        self.output.write('Master recieved '+repr(len(migrants))+' sets of migrants\n')
                        MIGEVENT+=1
                        if MIGEVENT>len(migrants):
                            MIGEVENT-=len(migrants)
                        nmigrants=migrants[MIGEVENT:]+migrants[:MIGEVENT]
                    else:
                        nmigrants=None
                    nmigrants = comm.scatter(nmigrants,root=0)
                    for one in nmigrants[0]:
                        pop.append(one)
            offspring = self.generation_set(pop)
            # Identify the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if ind.fitness==0]
            #Evaluate the individuals with invalid fitness
            self.output.write('\n--Evaluate Structures--\n')
            poorlist=[]
            logger.info('Evaluating fitness of structures')
            for i in range(len(invalid_ind)):
                ind = invalid_ind[i]
                outs = switches.fitness_switch([self,ind])
                self.output.write(outs[1])
                invalid_ind[i] = outs[0]
            pop.extend(invalid_ind)
            pop = self.generation_eval(pop)
            if self.convergence:
                self.overrideconvergence=True
            self.write()
        
        logger.info('Run algorithm stats')
        end_signal = self.algorithm_stats(self.population)
        return end_signal    
    
    def algorithm_parallel1(self):
        """Subprogram for running parallel version of GA
        Requires MPI4PY"""
        global logger
        comm = MPI.COMM_WORLD
        rank = MPI.COMM_WORLD.Get_rank()
        if 'MA' in self.debug:
            debug = True
        else:
            debug = False
        if rank==0:
            self.algorithm_initialize()
            logger.info('Beginning main algorithm loop')
        #Begin main algorithm loop
        self.convergence = False
        convergence=False
        while not convergence:
            if rank==0:
                pop = self.population
                offspring = self.generation_set(pop)
                # Identify the individuals with an invalid fitness
                invalid_ind = [ind for ind in offspring if ind.fitness==0]
                #Evaluate the individuals with invalid fitness
                self.output.write('\n--Evaluate Structures--\n')
                ntimes=int(math.ceil(float(len(invalid_ind))/float(comm.Get_size())))
                nadd=int(ntimes*comm.Get_size()-len(invalid_ind))
                maplist=[[] for n in range(ntimes)]
                strt=0
                for i in range(len(maplist)):
                    maplist[i]=[[self,indi] for indi in invalid_ind[strt:comm.Get_size()+strt]]
                    strt+=comm.Get_size()
                for i in range(nadd):
                    maplist[len(maplist)-1].append([None,None])
#                 for i in range(ntimes):
#                     for j in range(
            else:
                ntimes=None
#             worker = MPI.COMM_SELF.Spawn(cmd, None, 5)
# 
#             n  = array('i', [100])
#             worker.Bcast([n,MPI.INT], root=MPI.ROOT)
# 
#             pi = array('d', [0.0])
#             worker.Reduce(sendbuf=None,
#                           recvbuf=[pi, MPI.DOUBLE],
#                           op=MPI.SUM, root=MPI.ROOT)
#             pi = pi[0]
# 
#             worker.Disconnect()
#             ntimes = comm.bcast(ntimes,root=0)
            outs=[]
            for i in range(ntimes):
                if rank==0:
                    one=maplist[i]
                else:
                    one=None
                ind =comm.scatter(one,root=0)
                out = switches.fitness_switch(ind)
                outt = comm.gather(out,root=0)
                if rank==0:
                    outs.extend(outt)
            if rank==0:
                for i in range(len(invalid_ind)):
                    invalid_ind[i] = outs[i][0]
                for i in range(len(outs)):
                    self.output.write(outs[i][1])
                pop.extend(invalid_ind)
                pop = self.generation_eval(pop)
                self.write()
            convergence =comm.bcast(self.convergence, root=0)
        
        if rank==0:
            logger.info('Run algorithm stats')
            end_signal = self.algorithm_stats(self.population)
        else:
            end_signal = None
        end_signal = comm.bcast(end_signal, root=0)
        return end_signal    
    
    def algorithm_parallel(self):
        """Subprogram for running parallel version of GA
        Requires MPI4PY"""
        global logger
        comm = MPI.COMM_WORLD
        rank = MPI.COMM_WORLD.Get_rank()
        if 'MA' in self.debug:
            debug = True
        else:
            debug = False
        if rank==0:
            self.algorithm_initialize()
            logger.info('Beginning main algorithm loop')
        #Begin main algorithm loop
        self.convergence = False
        convergence=False
        while not convergence:
            if rank==0:
                pop = self.population
                offspring = self.generation_set(pop)
                # Identify the individuals with an invalid fitness
                invalid_ind = [ind for ind in offspring if ind.fitness==0]
                #Evaluate the individuals with invalid fitness
                self.output.write('\n--Evaluate Structures--\n')
                ntimes=int(math.ceil(float(len(invalid_ind))/float(comm.Get_size())))
                nadd=int(ntimes*comm.Get_size()-len(invalid_ind))
                maplist=[[] for n in range(ntimes)]
                strt=0
                for i in range(len(maplist)):
                    maplist[i]=[[self,indi] for indi in invalid_ind[strt:comm.Get_size()+strt]]
                    strt+=comm.Get_size()
                for i in range(nadd):
                    maplist[len(maplist)-1].append([None,None])
            else:
                ntimes=None
            ntimes = comm.bcast(ntimes,root=0)
            outs=[]
            for i in range(ntimes):
                if rank==0:
                    one=maplist[i]
                else:
                    one=None
                ind =comm.scatter(one,root=0)
                out = switches.fitness_switch(ind)
                outt = comm.gather(out,root=0)
                if rank==0:
                    outs.extend(outt)
            if rank==0:
                for i in range(len(invalid_ind)):
                    invalid_ind[i] = outs[i][0]
                for i in range(len(outs)):
                    self.output.write(outs[i][1])
                pop.extend(invalid_ind)
                pop = self.generation_eval(pop)
                self.write()
            convergence =comm.bcast(self.convergence, root=0)
        
        if rank==0:
            logger.info('Run algorithm stats')
            end_signal = self.algorithm_stats(self.population)
        else:
            end_signal = None
        end_signal = comm.bcast(end_signal, root=0)
        return end_signal    
    
    def algorithm_par_mp1(self):
        """Subprogram for running parallel version of GA
        Requires MPI4PY"""
        comm = MPI.COMM_WORLD
        rank = MPI.COMM_WORLD.Get_rank()
        if rank==0:
            if 'MA' in self.debug:
                debug = True
            else:
                debug = False
            self.algorithm_initialize()
        self.convergence = False
        convergence = False
        while not convergence:
            if rank==0:
                self.calc = tools.setup_calculator(self) #Set up calculator for atomic structures
                #Set up calculator for fixed region calculations
                if self.fixed_region:
                    self.static_calc = self.calc #May need to copy this
                    self.calc = tools.setup_fixed_region_calculator(self)
                pop = self.population
                offspring = self.generation_set(self,pop)
                # Identify the individuals with an invalid fitness
                invalid_ind = [ind for ind in offspring if ind.energy==0]
                #Evaluate the individuals with invalid fitness
                self.output.write('\n--Evaluate Structures--\n')
                proc_dist = int(comm.Get_size()/self.n_proc_eval)
                ntimes=int(math.ceil(float(len(invalid_ind))/float(proc_dist)))
                nadd=int(ntimes*proc_dist-len(invalid_ind))
                maplist=[[] for n in range(ntimes)]
                strt=0
                for i in range(len(maplist)):
                    maplist[i]=[[self,indi] for indi in invalid_ind[strt:proc_dist+strt]]
                    strt+=proc_dist
                for i in range(nadd):
                    maplist[len(maplist)-1].append([None,None])
                masterlist = [i*self.n_proc_eval for i in range(proc_dist)]
            else:
                ntimes=None
                masterlist = None
            ntimes = comm.bcast(ntimes,root=0)
            outs=[]
            for i in range(ntimes):
                if rank==0:
                    one=maplist[i]
                    for j in range(len(one)):
                        comm.send(ind, dest=1, tag=11)
                elif rank == 1:
                    ind = comm.recv(source=0, tag=11)
                else:
                    one=None
                ind =comm.scatter(one,root=0)
                out = switches.fitness_switch(ind)
            else:
                invalid_ind=[]
            poorlist = []
            for i in range(len(invalid_ind)):
                if self.fitness_scheme=='STEM_Cost':
                    if self.stem_coeff==None:
                        ind = invalid_ind.pop()
                        from MAST.structopt.tools.StemCalc import find_stem_coeff
                        outs = find_stem_coeff(self,ind)
                        ind = outs[1]
                        self.stem_coeff = outs[0]
                        self.output.write('stem_coeff Calculated to be: '+repr(self.stem_coeff)+'\n')
                        pop.append(ind)
                ind=invalid_ind[i]
                if 'MA' in self.debug: write_xyz(self.debugfile,ind[0],'Individual to fitness_switch')
                outs = switches.fitness_switch([self,ind])
                self.output.write(outs[1])
                invalid_ind[i]=outs[0]
                if invalid_ind[i].energy == float('inf'):
                    poorlist.append(i)
                    self.output.write('Removing infinite energy individual '+repr(ind.history_index)+'\n')
                elif invalid_ind[i].energy == float('nan'):
                    poorlist.append(i)
                    self.output.write('Removing nan energy individual '+repr(ind.history_index)+'\n')
                self.output.flush()
            if len(poorlist) != 0:
                poorlist.sort(reverse=True)
                for one in poorlist:
                    del invalid_ind[one]
            if rank==0:
                pop.extend(invalid_ind)
                pop = self.generation_eval(pop)
            convergence = comm.bcast(self.convergence, root=0)
        
        if rank==0:
            end_signal = self.algorithm_stats(self.population)
        else:
            end_signal = None
        end_signal = comm.bcast(end_signal, root=0)
        return end_signal    
    
    def algorithm_par_mp(self):
        """Subprogram for running parallel version of GA
        Requires MPI4PY"""
        global logger
        comm = MPI.COMM_WORLD
        rank = MPI.COMM_WORLD.Get_rank()
        if rank==0:
            if 'MA' in self.debug:
                debug = True
            else:
                debug = False
            self.algorithm_initialize()
        self.convergence = False
        convergence = False
        while not convergence:
            if rank==0:
                pop = self.population
                offspring = self.generation_set(self,pop)
                # Identify the individuals with an invalid fitness
                invalid_ind = [ind for ind in offspring if ind.energy==0]
                #Evaluate the individuals with invalid fitness
                self.output.write('\n--Evaluate Structures--\n')
            else:
                invalid_ind=[]
            for i in range(len(invalid_ind)):
                if self.fitness_scheme=='STEM_Cost':
                    if self.stem_coeff==None:
                        ind = invalid_ind.pop()
                        from MAST.structopt.tools.StemCalc import find_stem_coeff
                        outs = find_stem_coeff(self,ind)
                        ind = outs[1]
                        self.stem_coeff = outs[0]
                        self.output.write('stem_coeff Calculated to be: '+repr(self.stem_coeff)+'\n')
                        pop.append(ind)
                ind=invalid_ind[i]
                if 'MA' in self.debug: write_xyz(self.debugfile,ind[0],'Individual to fitness_switch')
                outs = switches.fitness_switch([self,ind])
                self.output.write(outs[1])
                invalid_ind[i]=outs[0]
                self.output.flush()
            if rank==0:
                pop.extend(invalid_ind)
                pop = self.generation_eval(pop)
                self.write()
            convergence = comm.bcast(self.convergence, root=0)
        
        if rank==0:
            end_signal = self.algorithm_stats(self.population)
        else:
            end_signal = None
        end_signal = comm.bcast(end_signal, root=0)
        return end_signal    
    
    def algorithm_serial(self):
        """Subprogram to run the optimizer in serial"""
        global logger
        self.algorithm_initialize()
        logger.info('Beginning main algorithm loop')
        #Begin main algorithm loop
        while not self.convergence:
            logger.info('Setup calculator for generation {0}'.format(self.generation))
            pop = self.population
            logger.info('Setup population')
            offspring = self.generation_set(pop)
            # Identify the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if ind.energy==0]
            #DEBUG: Write first invalid ind
            if 'MA' in self.debug:
                logger.info('Identified {0} structures with energy=0'.format(len(invalid_ind)))
                inp_out.write_xyz(self.debugfile,invalid_ind[0][0],\
                'First from Invalid_ind list '+repr(invalid_ind[0].energy))
                #DEBUG: Write first invalid ind in solid
                if self.structure=='Defect' or self.structure=='Surface':
                    sols = invalid_ind[0][0].copy()
                    sols.extend(invalid_ind[0].bulki)
                    inp_out.write_xyz(self.debugfile,sols,'First from Invalid-ind + Bulki')
                    sols = invalid_ind[0][0].copy()
                    sols.extend(invalid_ind[0].bulko)
                    inp_out.write_xyz(self.debugfile,sols,'First from Invalid-ind + Bulko')
            self.output.write('\n--Evaluate Structures--\n')
            logger.info('Evaluating fitness of structures')
            for i in range(len(invalid_ind)):
                ind = invalid_ind[i]
                if 'MA' in self.debug:
                    write_xyz(self.debugfile,ind[0],'Individual to fitness_switch')
                outs = switches.fitness_switch([self,ind])
                self.output.write(outs[1])
                invalid_ind[i] = outs[0]
            pop.extend(invalid_ind)
            logger.info('Eval Population')
            pop = self.generation_eval(pop)
            self.write()
            
        logger.info('Run algorithm stats')
        end_signal = self.algorithm_stats(self.population)
        return end_signal    
    
    def algorithm_stats(self,pop):
        self.output.write('\n----- Algorithm Stats -----\n')
        cxattempts = 0
        cxsuccess = 0
        for ats,sus in self.CXs:
            cxattempts+=ats
            cxsuccess+=sus
        mutslist = [[0,0] for one in self.mutation_options]
        for one in self.Muts:
            for i in range(len(mutslist)):
                mutslist[i][0]+=one[i][0]
                mutslist[i][1]+=one[i][1]
        self.output.write('Total Number of Evaluations : '+repr(sum(self.Evaluations))+'\n')
        self.output.write('Average Number of Evaluations per Generation : '+repr(
            float(sum(self.Evaluations))/float(len(self.Evaluations)))+'\n')
        ttseconds = self.Runtimes[-1]-self.Runtimes[0]
        deltats = []
        for i in range(1,len(self.Runtimes)):
            deltats.append(self.Runtimes[i]-self.Runtimes[i-1])
        maxs = [(ind,value) for ind,value in enumerate(deltats) if value==max(deltats)]
        maxtgen=maxs[0][0]-1
        maxts=maxs[0][1]
        mins = [(ind,value) for ind,value in enumerate(deltats) if value==min(deltats)]
        mintgen=mins[0][0]-1
        mints=mins[0][1]
        avgts = sum(deltats)/len(deltats)
        self.output.write('Total Length of GA run : '+tools.convert_time(ttseconds)+'\n')
        self.output.write('Average time per generation : '+tools.convert_time(avgts)+'\n')
        self.output.write('Maximum time for generation '+repr(maxtgen)+' : '+tools.convert_time(maxts)+'\n')
        self.output.write('Minimum time for generation '+repr(mintgen)+' : '+tools.convert_time(mints)+'\n')
        self.output.write('Attempted Crossovers: ' + repr(cxattempts)+'\n')
        self.output.write('Successful Crossovers: ' + repr(cxsuccess)+'\n')
        self.output.write('Mutations:\n')
        i=0
        for opt in self.mutation_options:
            self.output.write('    Attempted ' + opt + ' : ' + repr(mutslist[i][0]) + '\n')
            self.output.write('    Successful ' + opt + ' : ' + repr(mutslist[i][1]) + '\n')
            i+=1
        if self.best_inds_list:
            BESTS=tools.BestInds(pop,self.BESTS,self,writefile=True)
        self.close_output()
        end_signal='Genetic Algorithm Finished'
        return end_signal
    
    def check_pop(self, pop):
        # Gather all the energies/fitnesses
        if self.output_format=='totalenergy':
            complist = [ind.energy for ind in pop]
        elif self.output_format=='formation_energy':
            complist = []
            for ind in pop:
                solid=Atoms()
                solid.extend(ind[0])
                solid.extend(ind.bulki)
                energy=ind.energy
                for sym,c,m,u in self.atomlist:
                    nc=len([atm for atm in solid if atm.symbol==sym])
                    energy-= float(nc)*float(u)
                complist.append(energy)
        elif self.output_format=='formation_energy_per_int':
            complist = []
            for ind in pop:
                solid=Atoms()
                solid.extend(ind[0])
                solid.extend(ind.bulki)
                energy=ind.energy
                for sym,c,m,u in self.atomlist:
                    nc=len([atm for atm in solid if atm.symbol==sym])
                    energy-= float(nc)*float(u)
                energy=energy/self.natoms
                complist.append(energy)
        elif self.output_format=='formation_energy2':
            complist = [(ind.energy - ind.purebulkenpa*(ind[0].get_number_of_atoms()+ind.bulki.get_number_of_atoms()))/(ind[0].get_number_of_atoms()+ind.bulki.get_number_of_atoms()-ind.natomsbulk) for ind in pop]
        elif self.output_format=='energy_per_atom':
            complist = [ind.energy/(ind[0].get_number_of_atoms()+ind.bulki.get_number_of_atoms()) for ind in pop]
        else:
            complist = [ind.fitness for ind in pop]
    
        # Calcluate and print the Stats
        length = len(pop)
        mean = sum(complist) / length
        complist.sort()
        medium = complist[length/2]
        sum2 = sum(x*x for x in complist)
        std = abs(sum2 / length - mean**2)**0.5
        mine = min(complist)
        maxe = max(complist)
    
        self.output.write('\n----Stats----\n')
        self.output.write('  Min '+repr(mine)+'\n')
        self.output.write('  Max '+repr(maxe)+'\n')
        self.output.write('  Avg '+repr(mean)+'\n')
        self.output.write('  Medium '+repr(medium)+'\n')
        self.output.write('  Std '+repr(std)+'\n')
        self.output.write('  Genrep '+repr(self.genrep)+'\n')
        self.summary.write('{Gen} {Fitmin} {Fitavg} {Fitmed} {Fitmax} {Std} {time}\n'.format(
            Gen=str(self.generation),Fitmin=repr(mine),Fitavg=repr(mean),Fitmed=repr(medium),Fitmax=repr(maxe),
            Std=repr(std),time=repr(time.asctime( time.localtime(time.time()) ))))
    
        # Set new index values and write population
        index1 = 0
        for ind in pop:
            ind.index=index1
            index1+=1
        inp_out.write_pop(self,pop)
    
        if self.allenergyfile:
            for ind in pop:
                self.tenergyfile.write(repr(ind.energy)+' ')
            self.tenergyfile.write('\n')
    
        #Check Convergence of population based on fitness
        fitnesses = [ind.fitness for ind in pop]
        popmin = min(fitnesses)
        popmean = sum(fitnesses) / len(fitnesses)
        sum2 = sum(x*x for x in fitnesses)
        popstd = abs(sum2 / len(fitnesses) - popmean**2)**0.5
        convergence = False
        if self.convergence_scheme=='gen_rep_min':
            if self.generation < self.maxgen:
                if self.generation == 0:
                    self.minfit = popmin
                if abs(self.minfit - popmin) < self.tolerance:
                    self.genrep += 1
                else:
                    self.genrep = 0
                    self.minfit = popmin
                if self.genrep > self.reqrep:
                    self.convergence = True
            else:
                self.convergence = True
        elif self.convergence_scheme=='gen_rep_avg':
            if self.generation < self.maxgen:
                if self.generation == 0:
                    self.minfit = popmean
                if abs(self.minfit - popmean) < self.tolerance:
                    self.genrep += 1
                else:
                    self.genrep = 0
                    self.minfit = popmean
                if self.genrep > self.reqrep:
                    self.convergence = True
            else:
                self.convergence = True
        elif self.convergence_scheme=='std':
            if popstd < self.tolerance:
                self.convergence = True
        else:#self.convergence_scheme=='max_gen':
            if self.generation > self.maxgen:
                self.convergence = True
    
        #Flush output to files
        self.output.flush()
        self.summary.flush()
        for ind in self.files:
            ind.flush()
        if self.indiv_defect_write:
            for ind in self.ifiles:
                ind.flush()
        if self.genealogy: self.Genealogyfile.flush()
        if self.allenergyfile: self.tenergyfile.flush()
        if 'MA' in self.debug: self.debugfile.flush()
        if self.fingerprinting: 
            self.fpfile.flush()
            self.fpminfile.flush()
        return self

    def close_output(self):
        localtime = time.asctime( time.localtime(time.time()) )
        for ind in self.files:
            ind.close()
        self.output.write('Local time : '+repr(localtime)+'\n')
        self.output.write('End of Execution\n')
        self.summary.close()
        self.output.close()
        if self.genealogy: self.Genealogyfile.close()
        if self.allenergyfile: self.tenergyfile.close()
        if 'MA' in self.debug: self.debugfile.close()
        if self.fingerprinting: 
            self.fpfile.close()
            self.fpminfile.close()
    
    def generation_eval(self, pop):
        global logger
        emx = max(ind.energy for ind in pop)
        emn = min(ind.energy for ind in pop)
        for ind in pop:
            ind.tenergymx = emx
            ind.tenergymin = emn
        #DEBUG: Write relaxed individual
        if 'MA' in self.debug:
            if self.generation > 0: 
                inp_out.write_xyz(self.debugfile,pop[self.nindiv][0],\
                'First Relaxed Offspring '+repr(pop[self.nindiv-1].energy))    
                #DEBUG: Write relaxed ind in solid
                if self.structure=='Defect' or self.structure=='Surface':
                    inp_out.write_xyz(self.debugfile,pop[self.nindiv].bulki,\
                    'First Relaxed bulki '+repr(pop[self.nindiv-1].energy))    
                    sols = pop[self.nindiv][0].copy()
                    sols.extend(pop[self.nindiv].bulki)
                    inp_out.write_xyz(self.debugfile,sols,'First from Invalid-ind + Bulki '+\
                    repr(pop[self.nindiv].energy))
                    sols = pop[self.nindiv][0].copy()
                    sols.extend(pop[self.nindiv].bulko)
                    inp_out.write_xyz(self.debugfile,sols,\
                    'First from Invalid-ind + Bulko '+repr(pop[self.nindiv].energy))
        if self.generation==0:
            logger.info('Initializing Bests list')
            self.BESTS = list()
        if self.best_inds_list:
            self.BESTS = tools.BestInds(pop,self.BESTS,self,writefile=True)
        # Determine survival based on fitness predator
        if 'lambda,mu' not in self.algorithm_type:
            pop = tools.get_best(pop, len(pop))
        if self.fingerprinting:
            logger.info('Writing fingerprint files')
            for one in pop:
                self.fpfile.write(repr(fingerprinting.fingerprint_dist(
                    pop[0].fingerprint,one.fingerprint))+' '+repr(one.energy)+' ')
            self.fpfile.write('\n')
            self.fpminfile.write(repr(pop[0].fingerprint)+'\n')
            self.fpminfile.write(repr(pop[0].energy)+'\n')
        nevals = len(pop)/2
        if self.generation !=0:
            logger.info('Applying predator')
            pop = predator_switch(pop,self)
        else:
            self.genrep = 0
            self.minfit = 0
        # Evaluate population
        logger.info('Checking population for convergence')
        self.check_pop(pop)
        #Update general output tracking
        if self.generation !=0:
            histlist = []
            for ind in pop:
                histlist.append(ind.history_index)
            self.Runtimes.append(time.time())
            self.Evaluations.append(nevals)
            cxsuccess = 0
            mutsuccess = []
            for one in histlist:
                if '+' in one:
                    cxsuccess +=1
                if 'm' in one:
                    mutsuccess.append(one)
            self.CXs.append((self.cxattempts,cxsuccess))
            mutslist = [[0,0] for one in self.mutation_options]
            for one in mutsuccess:
                for two, opt in self.mutattempts:
                    if one==two:
                        index = [ind for ind,value in enumerate(
                            self.mutation_options) if value==opt][0]
                        mutslist[index][1]+=1
            for one,opt in self.mutattempts:
                index = [ind for ind,value in enumerate(
                self.mutation_options) if value==opt][0]
                mutslist[index][0]+=1
            self.Muts.append(mutslist)
            self.output.write('\n----- Generation Stats -----\n')
            self.output.write('Attempted Crossovers: ' + repr(self.cxattempts)+'\n')
            self.output.write('Successful Crossovers: ' + repr(cxsuccess)+'\n')
            self.output.write('Mutations:\n')
            i=0
            for opt in self.mutation_options:
                self.output.write('    Attempted ' + opt + ' : ' + repr(mutslist[i][0]) + '\n')
                self.output.write('    Successful ' + opt + ' : ' + repr(mutslist[i][1]) + '\n')
                i+=1
        self.generation += 1
        # Set new index values
        index1 = 0
        for ind in pop:
            ind.index = index1
            index1+=1
        index1 = 0
        if not self.convergence:
            try:
                self.calc.clean()
                if self.fixed_region:
                    self.static_calc.clean()
            except:
                pass
            if self.lammps_keep_files:
                try:
                    if self.parallel and ('Island_Method' not in self.algorithm_type):
                        nproc = MPI.COMM_WORLD.Get_size()
                        lmpfilepath = os.path.dirname(self.calc.tmp_dir)
                        for proc in range(nproc):
                            pth = os.path.join(lmpfilepath,'rank-{0}'.format(proc))
                            fls = [fl for fl in os.listdir(pth) if '.' not in fl]
                            for one in fls:
                                os.remove(pth+'/'+one)
                    else:
                        fls = [fl for fl in os.listdir(self.calc.tmp_dir) if '.' not in fl]
                        for one in fls:
                            os.remove(self.calc.tmp_dir+'/'+one)
                except Exception, e:
                    logger.error('Print issue in removing files {0}.'.format(e),exc_info=True)
                    print str(e)
                    pass
        self.population = pop
        return pop
    
    def generation_set(self,pop):
        global logger
        self.calc = tools.setup_calculator(self) #Set up calculator for atomic structures
        #Set up calculator for fixed region calculations
        if self.fixed_region:
            self.static_calc = self.calc #May need to copy this
            self.calc = tools.setup_fixed_region_calculator(self)
        self.output.write('\n-------- Generation '+repr(self.generation)+' --------\n')
        self.files[self.nindiv].write('Generation '+str(self.generation)+'\n')
        if len(pop) == 0:
            logger.info('Initializing structures')
            offspring = self.initialize_structures()
            self.population = offspring
        else:
            for i in range(len(pop)):
                # Reset History index
                pop[i].history_index=repr(pop[i].index)
            # Select the next generation individuals
            offspring = switches.selection_switch(pop, self.nindiv,
                        self.selection_scheme, self)
            # Clone the selected individuals
            offspring=[off1.duplicate() for off1 in offspring]
            # Apply crossover to the offspring
            self.output.write('\n--Applying Crossover--\n')
            cxattempts = 0
            for child1, child2 in zip(offspring[::2], offspring[1::2]):
                if random.random() < self.cxpb:
                    child1,child2 = switches.crossover_switch(child1, child2, self)
                    cxattempts+=2
            self.cxattempts=cxattempts
            #DEBUG: Write first child
            if 'MA' in self.debug: 
                inp_out.write_xyz(self.debugfile,offspring[0][0],'First Child '+
                    repr(offspring[0].history_index))
            # Apply mutation to the offspring
            self.output.write('\n--Applying Mutation--\n')
            mutattempts = []
            muts = []
            for mutant in offspring:
                if random.random() < self.mutpb:
                    if self.mutant_add:
                        mutant = mutant.duplicate()
                    mutant, optsel = switches.moves_switch(mutant,self)
                    mutattempts.append([mutant.history_index,optsel])
                    if self.mutant_add:
                        muts.append(mutant)
            if self.mutant_add:
                offspring.extend(muts)
            self.mutattempts=mutattempts
            #DEBUG: Write first offspring
            if 'MA' in self.debug: 
                inp_out.write_xyz(self.debugfile,muts[0][0],'First Mutant '+\
                repr(muts[0].history_index))
        if 'stem' in self.fitness_scheme:
            if self.stem_coeff==None:
                logger.info('Setting STEM coeff (alpha)')
                ind = offspring.pop()
                from MAST.structopt.tools.StemCalc import find_stem_coeff
                outs = find_stem_coeff(self,ind)
                ind = outs[1]
                ind.fitness = 0
                ind.energy = 0
                self.stem_coeff = outs[0]
                logger.info('STEM Coeff = {0}'.format(self.stem_coeff))
                self.output.write('stem_coeff Calculated to be: '+repr(self.stem_coeff)+'\n')
                offspring.append(ind)
                  
        return offspring
    
    def initialize_structures(self):
        global logger
        self.output.write('\n----Initialize Structures----\n')
        #self.logger.info('Initializing Structures')
        # Initialize Population - Generate a list of ncluster individuals
        # Set bulk and index atributes
        if self.restart:
            logger.info('Loading previous population')
            pop = generate.get_restart_population(self)
        else:
            logger.info('Generating new population')
            pop = generate.get_population(self)
        if 'MA' in self.debug: 
            inp_out.write_xyz(self.debugfile,pop[0][0],'First Generated Individual')
        #Use if concentration of interstitials is unknown
        if self.swaplist:
            mutlist=self.mutation_options
            self.mutation_options=['IntSwapLocal']
            for i in range(len(pop)):
                one = pop[i]
                one.swaplist=copy.deepcopy(swaplist)
                if random.random() < 0.25:
                    pop[i], opt = switches.moves_switch(one,self)
            self.mutation_options=mutlist
        # Write Initial structures to files
        self.output.write('\n---Starting Structures---\n')
        inp_out.write_pop(self, pop)
        #Print number of atoms to Summary file
        if self.structure=='Defect' or self.structure=='Surface':
            natstart=len(pop[0][0])+len(pop[0].bulki)
        else:
            natstart=len(pop[0][0])
        if not self.restart:
            if self.structure=='Defect':
                self.summary.write('Defect Run Pure Bulk Energy per Atom : '+repr(self.purebulkenpa)+'\n')
            else:
                self.summary.write(self.structure+' Run : '+repr(0)+'\n')
            self.summary.write('Natoms '+repr(natstart)+ '\n')
            #Print data headers to summary file
            self.summary.write('Generation Fitmin Fitavg Fitmedium Fitmax std time \n')
        offspring=pop
        #pop=[]
        #BESTS=[]
        return offspring
    
    def run(self):
        global logger
        cwd=os.getcwd()
        try:
            if self.parallel:
                MPI.COMM_WORLD.Set_errhandler(MPI.ERRORS_ARE_FATAL) 
                if 'Island_Method' in self.algorithm_type:
                    logger.info('Running Island Method algorithm')
                    self.algorithm_island()
                    if self.postprocessing:
                        logger.info('Running Post-processing')
                        rank = MPI.COMM_WORLD.Get_rank()
                        path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                        os.chdir(path)
                        if self.genealogytree:
                            pp.read_output(os.getcwd(),genealogytree=True,natoms=self.natoms,loggername=self.loggername)
                        else:
                            pp.read_output(os.getcwd(),genealogytree=False,natoms=self.natoms, loggername=self.loggername)
                        os.chdir(cwd)
                    if self.lattice_concentration:
                        if self.structure=='Defect':
                            logger.info('Running lattice concentration check')
                            rank = MPI.COMM_WORLD.Get_rank()
                            path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                            os.chdir(path)
                            if self.best_inds_list:
                                pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'Bests-'+self.filename+'.xyz'))
                            else:
                                pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'indiv00.xyz'))
                            os.chdir(cwd)
                elif 'MP_Eval' in self.algorithm_type:
                    logger.info('Running multiprocessor parallel algorithm')
                    self.algorithm_par_mp()
                    rank = MPI.COMM_WORLD.Get_rank()
                    if rank==0:
                        if self.postprocessing:
                            logger.info('Running Post-processing')
                            path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                            os.chdir(path)
                            if self.genealogytree:
                                pp.read_output(os.getcwd(),genealogytree=True,natoms=self.natoms,loggername=self.loggername)
                            else:
                                pp.read_output(os.getcwd(),genealogytree=False,natoms=self.natoms, loggername=self.loggername)
                            os.chdir(cwd)
                    if self.lattice_concentration:
                        if self.structure=='Defect':
                            logger.info('Running lattice concentration check')
                            path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                            os.chdir(path)
                            if self.best_inds_list:
                                pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'Bests-'+self.filename+'.xyz'))
                            else:
                                pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'indiv00.xyz'))
                            os.chdir(cwd)
                else:
                    logger.info('Running parallel algorithm')
                    self.algorithm_parallel()
                    rank = MPI.COMM_WORLD.Get_rank()
                    if rank==0:
                        if self.postprocessing:
                            logger.info('Running Post-processing')
                            path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                            os.chdir(path)
                            if self.genealogytree:
                                pp.read_output(os.getcwd(),genealogytree=True,natoms=self.natoms)
                            else:
                                pp.read_output(os.getcwd(),genealogytree=False,natoms=self.natoms)
                            os.chdir(cwd)
                        if self.lattice_concentration:
                            if self.structure=='Defect':
                                logger.info('Running lattice concentration check')
                                path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                                os.chdir(path)
                                if self.best_inds_list:
                                    pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'Bests-'+self.filename+'.xyz'))
                                else:
                                    pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'indiv00.xyz'))
                                os.chdir(cwd)
            else:
                logger.info('Running serial algorithm')
                self.algorithm_serial()
                rank = 0
                if self.postprocessing:
                    logger.info('Running Post-processing')
                    path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                    os.chdir(path)
                    if self.genealogytree:
                        pp.read_output(os.getcwd(),genealogytree=True,natoms=self.natoms)
                    else:
                        pp.read_output(os.getcwd(),genealogytree=False,natoms=self.natoms)
                    os.chdir(cwd)
                if self.lattice_concentration:
                    if self.structure=='Defect':
                        logger.info('Running lattice concentration check')
                        path = os.path.join(os.getcwd(), '{0}-rank{1}'.format(self.filename,rank))
                        os.chdir(path)
                        if self.best_inds_list:
                            pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'Bests-'+self.filename+'.xyz'))
                        else:
                            pp.get_lattice_concentration(os.path.join(os.getcwd(),'Bulkfile.xyz'),os.path.join(os.getcwd(),'indiv00.xyz'))
                        os.chdir(cwd)
        except Exception, e:
            logger.error('Error in execution: {0}'.format(e),exc_info=True)
            print '********ERROR IN EXECUTION********'
            print str(e)
            print 'CLEANING UP FILES'
            try:
                self.close_output()
            except:
                pass
            print 'EXITING PROGRAM'
    
    def read(self,optfile):
        parameters = inp_out.read_parameter_input(optfile,True)
        self.__dict__.update(parameters)
        outdict = inp_out.restart_output(self)
        self.__dict__.update(outdict)
        poplist = []
        for indfile in self.population:
            ind = inp_out.read_individual(indfile)
            poplist.append(ind)
        self.population = poplist
        bestlist = []
        for bestfile in self.BESTS:
            ind = inp_out.read_individual(bestfile)
            bestlist.append(ind)
        self.BESTS = bestlist
        self.restart = True
        return self
        
    def write(self,filename=None, restart=True):
        if filename:
            inp_out.write_optimizer(self, filename, restart)
        else:
            inp_out.write_optimizer(self, self.optimizerfile, restart)
        return
    
if __name__ == "__main__":
    import sys
    input = sys.argv[1]
    A=Optimizer(input)
    A.run()



##############################################################
# This code is part of the MAterials Simulation Toolkit (MAST)
# 
# Maintainer: Tam Mayeshiba
# Last updated: 2014-04-25
##############################################################
from custodian.vasp import handlers
from pymatgen.core.structure import Structure
import logging
import inspect
import os
from MAST.ingredients.errorhandler import BaseError
from MAST.ingredients.errorhandler import VaspError
from MAST.ingredients.errorhandler import masterrorhandlers
class VaspNEBError(BaseError):
    """VASP NEB error-handling functions (wraps custodian)
    """
    def __init__(self, **kwargs):
        allowed_keys = {
            'name' : (str, str(), 'Name of directory'),
            'program_keys': (dict, dict(), 'Dictionary of program keywords'),
            'structure': (Structure, None, 'Pymatgen Structure object')
            }
        BaseError.__init__(self, allowed_keys, **kwargs)


    def set_handler_inputs(self):
        """We may want to override the files that are being checked, which are
            usually vasp.out or vasprun.xml.
            Use key as error handler class name (like 'VaspErrorHandler').
            Use mast_skip to skip the error.
        """
        handler_input_d=dict()
        numim = self.keywords['program_keys']['mast_neb_settings']['images']
        numim = int(numim)
        warchlist = list()
        marchlist = list()
        clist = list()
        plist = list()
        for imct in range(1, numim+1):
            imstr = str(imct).zfill(2)
            warchlist.append("%s/OUTCAR" % imstr)
            warchlist.append("%s/OSZICAR" % imstr)
            warchlist.append("%s/CONTCAR" % imstr)
            warchlist.append("%s/POSCAR" % imstr)
            warchlist.append("%s/XDATCAR" % imstr)
            marchlist.append("%s/OUTCAR" % imstr)
            marchlist.append("%s/OSZICAR" % imstr)
            marchlist.append("%s/CONTCAR" % imstr)
            marchlist.append("%s/XDATCAR" % imstr)
            clist.append("%s/CONTCAR" % imstr)
            plist.append("%s/POSCAR" % imstr)
        handler_input_d['MASTWalltimeErrorHandler']=[self.keywords['name'],warchlist, clist, plist]
        handler_input_d['MASTMemoryErrorHandler']=[self.keywords['name'],self.keywords,marchlist]
        handler_input_d['VaspReachedNSWErrorHandler']=[self.keywords['name'],warchlist, clist, plist]

        return handler_input_d



import os
try:
    from mpi4py import MPI
except:
    pass
from MAST.structopt.inp_out.write_individual import write_individual

def write_optimizer(Optimizer, optfile, restart=True):
    """Function to write out an Optimizer class object
    Input:
        Optimizer = Optimizer class object to be written out
        optfile = String or fileobject that Optimizer class is written to
    Output:
        No output returned but optimizer object is written to specified file
    """
    if isinstance(optfile, str):
        optfile=open(optfile, 'w')
    convergencesettings = ['convergence_scheme',
        'maxgen',
        'reqrep',
        'convergence',
        'overrideconvergence']
    parametersettings = ['parallel',
        'filename',
        'loggername',
        'atomlist',
        'structure',
        'natoms',
        'optimizer_type',
        'nindiv',
        'genealogy',
        'output_format',
        'allenergyfile',
        'best_inds_list',
        'number_of_bests',
        'indiv_defect_write',
        'vacancy_output',
        'lattice_concentration',
        'postprocessing',
        'genealogytree',
        'seed',
        'forcing',
        'debug',
        'algorithm_type',
        'migration_intervals',
        'migration_percent',
        'fingerprinting',
        'fpbin',
        'fpcutoff',
        'bulkfp',
        'fixed_region',
        'rattle_atoms',
        'constrain_position',
        'restart_ints',
        'restart',
        'r_ab',
        'size',
        'generate_flag',
        'sf',
        'supercell',
        'solidfile',
        'solidcell',
        'evalsolid',
        'finddefects',
        'trackvacs',
        'trackswaps',
        'random_loc_start',
        'random_vac_start',
        'purebulkenpa',
        'natomsbulk',
        'surfacefile',
        'surftopthick',
        'surfacecell',
        'cell_shape_options',
        'alloy',
        'calc_method',
        'vaspcalc',
        'pair_style',
        'bopcutoff',
        'buckcutoff',
        'buckparameters',
        'ps_name',
        'pair_coeff',
        'ps_other',
        'pot_file',
        'lammps_keep_files',
        'lammps_thermo_steps',
        'lammps_min',
        'lammps_min_style',
        'large_box_size',
        'ase_min',
        'ase_min_fmax',
        'ase_min_maxsteps',
        'cxpb',
        'cx_scheme',
        'selection_scheme',
        'mutpb',
        'mutation_options',
        'bh_temp',
        'bh_steps',
        'mutant_add',
        'quench_max_temp',
        'quench_min_temp',
        'quench_n_steps_1',
        'quench_n_steps_2',
        'quench_step_size',
        'isolate_mutation',
        'fitness_scheme',
        'energy_cutoff_factor',
        'stem_coeff',
        'stem_parameters' ,
        'stem_keep_files',
        'constrain_swaps',
        'swaplist',
        'natural_selection_scheme',
        'tournsize',
        'fusslimit',
        'metropolis_temp',
        'tolerance',
        'predator',
        'adaptbegin',
        'adaptmultiplier',
        'demin',
        'mark',
        'restart_optimizer']
    attributelist = ['genrep',
        'minfit',
        'generation',
        'Runtimes',
        'Evaluations',
        'CXs',
        'Muts',
        'cxattempts',
        'mutattempts',
        'optimizerfile']
    Optimizer.restart_optimizer = restart
    #Write Optimizer convergence parameters
    for one in convergencesettings:
        try:
            optpar = eval('Optimizer.{0}'.format(one))
            if isinstance(optpar,str):
                if '\n' in optpar:
                    optfile.write("'{0}':{1},\n".format(one,repr(optpar)))
                else:
                    optfile.write("'{0}':'{1}',\n".format(one,optpar))
            else:
                optfile.write("'{0}':{1},\n".format(one,optpar))
        except:
            optfile.write("'{0}':{1},\n".format(one,None))
            print 'Cannot write parameter: {0}'.format(one)
    #Write Optimizer parameters
    for one in parametersettings:
        try:
            optpar = eval('Optimizer.{0}'.format(one))
            if isinstance(optpar,str):
                if '\n' in optpar:
                    optfile.write("'{0}':{1},\n".format(one,repr(optpar)))
                else:
                    optfile.write("'{0}':'{1}',\n".format(one,optpar))
            else:
                optfile.write("'{0}':{1},\n".format(one,optpar))
        except:
            optfile.write("'{0}':{1},\n".format(one,None))
            print 'Cannot write parameter: {0}'.format(one)
    #Write Optimizer attributes
    for one in attributelist:
        try:
            optpar = eval('Optimizer.{0}'.format(one))
            if isinstance(optpar,str):
                if '\n' in optpar:
                    optifile.write("'{0}':{1},\n".format(one,repr(optpar)))
                else:
                    optfile.write("'{0}':'{1}',\n".format(one,optpar))
            else:
                optfile.write("'{0}':{1},\n".format(one,optpar))
        except:
            optfile.write("'{0}':{1},\n".format(one,None))
            print 'Cannot write attribute: {0}'.format(one)
    #Write Optimizer output files
    flist = []
    try:
        for i in range(len(Optimizer.files)):
            flist.append(Optimizer.files[i].name)
    except:
        flist = []
    optfile.write("'files':{0},\n".format(flist))
    iflist = []
    try:
        if Optimizer.ifiles:
            for i in range(len(Optimizer.ifiles)):
                iflist.append(Optimizer.ifiles[i].name)
        else:
            iflist = None
    except:
        iflist = None
    optfile.write("'ifiles':{0},\n".format(iflist))
    outputfiles = ['tenergyfile','fpfile','fpminfile','debugfile','Genealogyfile','output','summary']
    for one in outputfiles:
        try:
            optfile.write("'{0}':'{1}',\n".format(one,eval('Optimizer.{0}.name'.format(one))))
        except:
            optfile.write("'{0}':{1},\n".format(one,None))
            #print 'Cannot write filename: {0}'.format(one)
    #Write population and bests
    try:
        rank = MPI.COMM_WORLD.Get_rank()
    except:
        rank = 0
    if (len(Optimizer.population) > 0) and (len(Optimizer.BESTS) > 0):
        if isinstance(Optimizer.population[0], str):
            optfile.write("'population':{0},\n".format(Optimizer.population))
            optfile.write("'BESTS':{0}".format(Optimizer.BESTS))
        else:
            fpath = os.path.join(os.getcwd(), '{0}-rank{1}'.format(Optimizer.filename,rank))
            path = os.path.join(fpath,'Restart-files')
            if not os.path.exists(path):
                os.mkdir(path)
            popfiles = []
            for index in range(len(Optimizer.population)):
                fname = os.path.join(path,'Reload-indiv{0:02d}.txt'.format(index))
                write_individual(Optimizer.population[index],fname)
                popfiles.append(fname)
            optfile.write("'population':{0},\n".format(popfiles))
            bestfiles = []
            for index in range(len(Optimizer.BESTS)):
                fname = os.path.join(path,'Reload-bests{0:02d}.txt'.format(index))
                write_individual(Optimizer.BESTS[index],fname)
                bestfiles.append(fname)
            optfile.write("'BESTS':{0}".format(bestfiles))
    else:
        optfile.write("'population':{0},\n".format([]))
        optfile.write("'BESTS':{0}".format([]))
    optfile.close()
    return


import random
import numpy
from MAST.structopt.tools.find_defects import find_defects
from ase import Atom, Atoms
from ase.calculators.neighborlist import NeighborList

def lattice_alteration_nn(indiv, Optimizer):
    """Move function to perform random move along random axis for nearest neighbor distance to random atoms
    Inputs:
        indiv = Individual class object to be altered
        Optimizer = Optimizer class object with needed parameters
    Outputs:
        indiv = Altered Individual class object
    """
    if 'MU' in Optimizer.debug:
        debug = True
    else:
        debug = False
    if Optimizer.structure=='Defect':
        if Optimizer.isolate_mutation:
            indc,indb,vacant,swaps,stro = find_defects(indiv[0],Optimizer.solidbulk,0)
            ind = indc.copy()
            ind.extend(indb)
        else:
            ind=indiv[0].copy()
            indc=indiv[0].copy()
    else:
        ind=indiv[0].copy()
        indc=indiv[0].copy()
    if len(indc) != 0:
        ctoff1 = [1.0 for one in ind]
        nl = NeighborList(ctoff1, bothways=True, self_interaction=False)
        nl.update(ind)
        try:
            natomsmove=random.randint(1,len(indc)/2)
        except ValueError:
            natomsmove=1
        passn=0
        for count in range(natomsmove):
            try:
                indexmv = random.choice([i for i in range(len(indc))])
                indices, offsets = nl.get_neighbors(indexmv)
                nns = Atoms()
                nns.append(ind[indexmv])
                for index, d in zip(indices,offsets):
                    index = int(index)
                    pos = ind[index].position + numpy.dot(d,ind.get_cell())
                    nns.append(Atom(symbol=ind[index].symbol, position=pos))
                dist = [nns.get_distance(0,i) for i in range(1, len(nns))]
                r = sum(dist)/len(dist)
                dir = random.choice([[1,0,0],[-1,0,0],[0,1,0],[0,-1,0],[0,0,1],[0,0,-1]])
                ind[indexmv].position += [i*r for i in dir]
            except:
                passn+=1
        indiv[0]=ind.copy()
    else:
        natomsmove=0
        passn=0
    Optimizer.output.write('Lattice Alteration NN Mutation performed on individual\n')
    Optimizer.output.write('Index = '+repr(indiv.index)+'\n')
    natomsmove-=passn
    Optimizer.output.write('Number of atoms moved = '+repr(natomsmove)+'\n')
    Optimizer.output.write(repr(indiv[0])+'\n')
    muttype='LANN'+repr(natomsmove)
    if indiv.energy==0:
        indiv.history_index=indiv.history_index+'m'+muttype
    else:
        indiv.history_index=repr(indiv.index)+'m'+muttype
    return indiv


'''General functions for moves'''
from ase_minimization import *
# from atoms_add import *
# from atoms_remove import *
# from basin_hop_la import *
# from basin_hop_permute import *
# from basin_hop_ra_atoms import *
# from basin_hop_rattle import *
# from basin_hop_rotate import *
# from cell_relax_lammps import *
# from cell_shape import *
# from lattice_alteration_crystal import *
# from lattice_alteration_group import *
# from lattice_alteration_nn import *
# from lattice_alteration_rdrd import *
# from lattice_alteration import *
# from move_la import *
# from permutation_bulk import *
# from permutation_crystal_multi import *
# from permutation_crystal import *
# from permutation import *
# from quench import *
# from random_replacement import *
# from rattle import *
# from rotation_geo import *
# from rotation import *
# from scale_size import *
# from swap_int_local import *
# from swap_int import *
# from swap_vacancy import *
# from swap import *
# from zp_rotation_fixed import *
# from zp_rotation import *


##############################################################
# This code is part of the MAterials Simulation Toolkit (MAST)
# 
# Maintainer: Tam Mayeshiba
# Last updated: 2014-04-25
##############################################################
import os
import time
import shutil
import logging
from MAST.utility import MASTError
from MAST.utility import dirutil
from MAST.utility import loggerutils
from MAST.parsers.inputparser import InputParser
from MAST.recipe.recipesetup import RecipeSetup
from MAST.utility import MASTFile

class MASTmon(object):
    """The MAST monitor runs on a submission node and checks
        the status of each recipe in the MAST_SCRATCH directory.
        Attributes:
            self.scratch <str>: MAST_SCRATCH
            self._ARCHIVE <str>: MAST_ARCHIVE
            self.logger <logging logger>
    """ 
    def __init__(self):

        self.scratch = dirutil.get_mast_scratch_path()
        self._ARCHIVE = dirutil.get_mast_archive_path()
        self.make_directories() 
        self.logger = logging.getLogger('mastmon')
        self.logger = loggerutils.add_handler_for_control(self.logger)
        self.logger.info("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
        self.logger.info("\nMAST monitor started at %s.\n" % time.asctime())
        self.logger.info("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
        self.run(1)

    def make_directories(self):
        """Attempt to make scratch and archive directories
            if they do not exist.
        """
        try:
            if not os.path.exists(self.scratch):
                os.makedirs(self.scratch)
            if not os.path.exists(self._ARCHIVE):
                os.makedirs(self._ARCHIVE)
        except:
            raise MASTError(self.__class__.__name__,
                    "Error making directory for MASTmon and completed recipes")

    def check_recipe_dir(self, fulldir, verbose):
        """Check a recipe directory.
            Args:
                fulldir <str>: full path of recipe directory
                verbose <int>: verbosity
        """
        shortdir = os.path.basename(fulldir) #only the recipe directory name
        if not os.path.exists(fulldir):
            raise MASTError(self.__class__.__name__, "No recipe directory at %s" % fulldir)
        if os.path.exists(os.path.join(fulldir, "MAST_SKIP")):
            self.logger.warning("Skipping recipe %s due to the presence of a MAST_SKIP file in the recipe directory." % shortdir)
            return
        if os.path.exists(os.path.join(fulldir, "MAST_ERROR")):
            self.logger.error("ATTENTION!: Skipping recipe %s due to the presence of a MAST_ERROR file in the recipe directory." % shortdir)
            return
        self.logger.info("--------------------------------")
        self.logger.info("Processing recipe %s" % shortdir)
        self.logger.info("--------------------------------")
        os.chdir(fulldir) #need to change directories in order to submit jobs?
        myipparser = InputParser(inputfile=os.path.join(fulldir, 'input.inp'))
        myinputoptions = myipparser.parse()
        rsetup = RecipeSetup(recipeFile=os.path.join(fulldir,'personal_recipe.txt'),
                inputOptions=myinputoptions,
                structure=myinputoptions.get_item('structure','structure'),
                workingDirectory=fulldir)
        recipe_plan_obj = rsetup.start()
        recipe_plan_obj.get_statuses_from_file()
        try:
            recipe_plan_obj.check_recipe_status(verbose)
        except Exception:
            import sys,traceback
            ex_type, ex, trbck = sys.exc_info()
            errortext = traceback.print_tb(trbck)
            del trbck
            raise MASTError(self.__class__.__name__,"Error in recipe %s as follows: %s %s %s" % (shortdir, ex_type, ex, errortext))
        os.chdir(self.scratch)
        if recipe_plan_obj.status == "C":
            shutil.move(fulldir, self._ARCHIVE)
            summarypath = "%s/%s/SUMMARY.txt" % (self._ARCHIVE, shortdir)
            if os.path.isfile(summarypath):
                self.logger.info("Recipe %s completed." % shortdir)
                self.logger.info("SUMMARY.txt below:")
                summarytext = MASTFile(summarypath)
                for myline in summarytext.data:
                    self.logger.info(myline.strip())
        self.logger.info("-----------------------------")
        self.logger.info("Recipe %s processed." % shortdir)
        self.logger.info("-----------------------------")


    def run(self, verbose=0):
        """Run the MAST monitor.
        """
        curdir = os.getcwd()
        try:
            os.chdir(self.scratch)    
        except:
            os.chdir(curdir)
            errorstr = "Could not change directories to MAST_SCRATCH at %s" % self.scratch
            raise MASTError(self.__class__.__name__, errorstr)
        
        #dirutil.lock_directory(self.scratch, 1) # Wait 5 seconds
        #Directory is now locked by mast initially, but gets
        #unlocked at the end of the mastmon run.
        
        recipe_dirs = dirutil.walkdirs(self.scratch,1,1)
        if verbose == 1:
            self.logger.info("================================")
            self.logger.info("Recipe directories:")
            for recipe_dir in recipe_dirs:
                self.logger.info(recipe_dir)
            self.logger.info("================================")

        for recipe_dir in recipe_dirs:
            self.check_recipe_dir(recipe_dir, verbose)
                
        dirutil.unlock_directory(self.scratch) #unlock directory
        os.chdir(curdir)



##############################################################
# This code is part of the MAterials Simulation Toolkit (MAST)
# 
# Maintainer: Tam Mayeshiba
# Last updated: 2014-04-25
##############################################################
import os
import sys
import time
from MAST.utility import fileutil
from MAST.utility import MASTFile
from MAST.utility import dirutil
def get_citations(recipedir):
    """Use mast_exec lines from the recipe directory in order
        to assign a CITATIONS section to the SUMMARY.txt file.
        Args:
            recipedir <str>: Recipe directory
        Returns:
            linelist <list of str>: Lines to add to SUMMARY.txt
    """
    mylist=list()
    mylist=['mast','pymatgen','spglib'] #need to use; removed python as standard
    myinput = os.path.join(recipedir, "input.inp")
    mastexecs = fileutil.grepme(myinput, "mast_exec")
    for mastexec in mastexecs:
        if 'vasp' in mastexec.lower():
            mylist.append('vasp')
            mastxcs = fileutil.grepme(myinput, "mast_xc")
            if len(mastxcs) >= 1:
                mylist.append('vasp_pps')
                for mastxc in mastxcs:
                    if ("pw" in mastxc.lower()) or ("paw" in mastxc.lower()):
                        mylist.append('vasp_paw')
            mastprogs = fileutil.grepme(myinput, "mast_program")
            if len(mastprogs) >= 1:
                for mastprog in mastprogs:
                    if ("vasp_neb" in mastprog.lower()):
                        mylist.append('vaspneb')
            if "tst" in mastexec.lower():
                mylist.append('vaspneb')
        #if 'phon' in mastexec.lower():
        #    mylist.append('phon')
        if 'structopt' in mastexec.lower():
            mylist.append('structopt')
    citationpath = os.path.join(dirutil.get_mast_control_path(),"citations")
    citationfiles = os.listdir(citationpath)
    linelist = list()
    uniquelist=list()
    for mylistitem in mylist:
        if not mylistitem in uniquelist:
            uniquelist.append(mylistitem)
    #print "UNIQUELIST: ", uniquelist
    for listitem in uniquelist:
        for citationfile in citationfiles:
            if citationfile[:-3] == listitem: #all except number
                cfpath = os.path.join(citationpath, citationfile)
                cffile = MASTFile(cfpath)
                linelist.extend(cffile.data)
    return linelist



# parsers/__init__.py

from MAST.parsers.inputparser import InputParser
from MAST.parsers.inputpythoncreator import InputPythonCreator
from MAST.parsers.indeploopinputparser import IndepLoopInputParser
from MAST.parsers.recipetemplateparser import RecipeTemplateParser



import pymatgen
import os
import shutil
import sys
import unittest
import time
import filecmp
from filecmp import dircmp
from nose import SkipTest

from MAST.parsers import InputParser
from MAST.parsers import InputPythonCreator
from MAST.utility import MASTError
from MAST.utility import MASTFile
from MAST.utility.dirutil import *

class TestInput(unittest.TestCase):
    def setUp(self):
        scripts=get_mast_install_path()
        os.chdir(os.path.join(scripts,'test','input_test'))
    
    def tearDown(self):
        if os.path.isfile("input_python_created"):
            os.remove("input_python_created")

    def test_inputparser(self):
        raise SkipTest
        myip = InputParser(inputfile='mast.inp')
        myoptions=myip.parse()
        print myoptions
        self.assertEqual(True,True)

    def test_inputpythoncreator(self):
        raise SkipTest
        myip = InputParser(inputfile='mast.inp')
        myoptions=myip.parse()
        myipc = InputPythonCreator(input_options=myoptions) 
        mylines=myipc.print_input_options()
        myfile = MASTFile()
        myfile.data = mylines
        myfile.to_file("./input_python_created")
        #print mylines
        self.assertTrue(filecmp.cmp("input_python_created","test_input_python_created"))




import random
import numpy
from ase import Atom, Atoms
from MAST.structopt.tools.find_defects import find_defects
from MAST.structopt.tools.position_average import position_average
from MAST.structopt.tools.shift_atoms import shift_atoms
from MAST.structopt.inp_out import write_xyz

def rotct_defect(ind1, ind2, Optimizer):
    """Rotate atoms cut and splice
    Translates atoms to center of positions first
    Rotates atoms randomly around center of mass and cuts with xy plane
    Maintains number of atoms
    Maintains concentration of atoms
    Returns individuals to standard positions at end (un-rotates)
    """
    if 'CX' in Optimizer.debug:
        debug = True
    else:
        debug = False
    Optimizer.output.write('Rotate Cut/Splice Cx for defects between individual '+repr(ind1.index)+' and individual '+repr(ind2.index)+'\n')
    
    #Perserve starting conditions of individual
    indi1 = ind1[0].copy()
    indi2 =ind2[0].copy()
    #Translate individuals so COP is at (0,0,0)
    if Optimizer.structure=='Defect':
        #Identify center of positions for defect structure
        indi1c,indi1b,vacant1,swap1,stro1 = find_defects(indi1,Optimizer.solidbulk,0)
        com1 = position_average(indi1c)
        indi1 = shift_atoms(indi1, com1)
        trans = [-p for p in numpy.maximum.reduce(indi1.get_cell())]
        indi1.translate(trans)
        #Do the same for second individual
        indi2c,indi2b,vacant2,swap2,stro2 = find_defects(indi2,Optimizer.solidbulk,0)
        com2 = position_average(indi2c)
        indi2 = shift_atoms(indi2, com2)
        indi2.translate(trans)
    else:
        com1 = indi1.get_center_of_mass()
        indi1.translate(-1*com1)
        com2 = indi2.get_center_of_mass()
        indi2.translate(-1*com2)
    #Select random axis, random angle, and random position and rotate individuals
    n=0
    while n<10:
        rax = random.choice(['x', '-x','y','-y','z','-z'])
        rang = random.random()*90
        indi1.rotate(rax,a=rang,center=[0,0,0],rotate_cell=False)
        #Search for atoms in individual 1 that are above the xy plane
        group1 = Atoms(cell=ind1[0].get_cell(),pbc=ind1[0].get_pbc())
        indices1=[]
        for one in indi1:
            if one.position[2] >= 0:
                group1.append(one)
                indices1.append(one.index)
        if len(group1) > 2 and len(group1) < len(indi1):
            break
        else:
            n+=1
            indi1.rotate(rax,a=-1*rang,center=[0,0,0], rotate_cell=False)
    indi2.rotate(rax,a=rang,center=[0,0,0],rotate_cell=False)
    if debug: 
        print 'Group1 size = ', len(group1)
        print 'Position = ', [0,0,0]
        print 'Angle = ', rang
        print 'Axis = ', rax
        print 'Number of tries = ', n+1
    if len(group1) != 0:
        #Apply concentration forcing if needed
        group2 = Atoms(cell=ind2[0].get_cell(),pbc=ind2[0].get_pbc())
        indices2 = []
        dellist = []
        for one in indi2:
            if one.position[2] >= 0:
                group2.append(one)
                indices2.append(one.index)
        if Optimizer.forcing=='Concentration':
            symlist = list(set(indi1.get_chemical_symbols()))
            seplist = [[atm for atm in group2 if atm.symbol == sym] for sym in symlist]
            group2n=Atoms(cell=group2.get_cell(),pbc=group2.get_pbc())
            indices2n = []
            dellist = []
            for one in group1:
                sym1=one.symbol
                listpos=[i for i,s in enumerate(symlist) if s==sym1][0]
                if len(seplist[listpos]) > 0:
                    pos = random.choice(range(len(seplist[listpos])))
                    group2n.append(seplist[listpos][pos])
                    indices2n.append(indices2[seplist[listpos][pos].index])
                    del seplist[listpos][pos]
                else:
                    dellist.append(one.index)
            if len(dellist) != 0:
                dellist.sort(reverse=True)
                for one in dellist:
                    del group1[one]
                    del indices1[one]
            indices2=indices2n
            group2=group2n.copy()
        else:
            dellist = []
            while len(group2) < len(group1)-len(dellist):
                #Too many atoms in group 1
                dellist.append(random.choice(group1).index)
            if len(dellist) != 0:
                dellist.sort(reverse=True)
                for one in dellist:
                    del group1[one]
                    del indices1[one]
            dellist = []
            while len(group1) < len(group2)-len(dellist):
                #Too many atoms in group 2
                dellist.append(random.choice(group2).index)
            if len(dellist) != 0:
                dellist.sort(reverse=True)
                for one in dellist:
                    del group2[one]
                    del indices2[one]
    
        other2 = Atoms(cell=ind2[0].get_cell(),pbc=ind2[0].get_pbc())
        for one in indi2:
            if one.index not in indices2:
                other2.append(one)
        other1 = Atoms(cell=ind1[0].get_cell(),pbc=ind1[0].get_pbc())
        for one in indi1:
            if one.index not in indices1:
                other1.append(one)
        indi1 = group2.copy()
        indi1.extend(other1)
        indi2 = group1.copy()
        indi2.extend(other2)
    
        #DEBUG: Write crossover to file
        if debug: 
            write_xyz(Optimizer.debugfile, group1,'group1')
            write_xyz(Optimizer.debugfile, other1,'other1')
            write_xyz(Optimizer.debugfile, group2,'group2')
            write_xyz(Optimizer.debugfile, other2,'other2')
            print 'Length of group1 = ',len(group1),'Length of group2',len(group2)
    
        #DEBUG: Check structure of atoms exchanged
        for sym,c,m,u in Optimizer.atomlist:
            nc=len([atm for atm in indi1 if atm.symbol==sym])
            Optimizer.output.write('CX ROTCT_Defect: Individual 1 contains '+repr(nc)+' '+repr(sym)+' atoms\n')
            nc=len([atm for atm in indi2 if atm.symbol==sym])
            Optimizer.output.write('CX ROTCT_Defect: Individual 2 contains '+repr(nc)+' '+repr(sym)+' atoms\n')
        if Optimizer.forcing !='Concentration':
            for i in range(len(Optimizer.atomlist)):
                atms1=[inds for inds in indi1 if inds.symbol==Optimizer.atomlist[i][0]]
                atms2=[inds for inds in indi2 if inds.symbol==Optimizer.atomlist[i][0]]
                if len(atms1)==0:
                    if len(atms2)==0:
                        indi1[random.randint(0,len(indi1)-1)].symbol==Optimizer.atomlist[i][0]
                        indi2[random.randint(0,len(indi2)-1)].symbol==Optimizer.atomlist[i][0]
                    else:
                        indi1.append(atms2[random.randint(0,len(atms2)-1)])
                        indi1.pop(random.randint(0,len(indi1)-2))
                else:
                    if len(atms2)==0:
                        indi2.append(atms1[random.randint(0,len(atms1)-1)])
                        indi2.pop(random.randint(0,len(indi2)-2))	
        indi1.rotate(rax,a=-1*rang,center=[0,0,0], rotate_cell=False)
        indi2.rotate(rax,a=-1*rang,center=[0,0,0], rotate_cell=False)
        if Optimizer.structure=='Defect':
            trans = [-p for p in trans]
            indi1.translate(trans)
            indi2.translate(trans)
            indi1 = shift_atoms(indi1, [-p for p in com1])
            indi2 = shift_atoms(indi2, [-p for p in com2])
        else:
            indi1.translate(com1)
            indi2.translate(com2)
        
        #DEBUG: Check structure and number of atoms in crystal
        if Optimizer.structure=='Defect':
            solid1=Atoms()
            solid1.extend(indi1)
            solid1.extend(ind1.bulki)
            solid2=Atoms()
            solid2.extend(indi2)
            solid2.extend(ind2.bulki)
            for sym,c,m,u in Optimizer.atomlist:
                nc=len([atm for atm in solid1 if atm.symbol==sym])
                Optimizer.output.write('CX ROTCT_Defect: Defect 1 configuration contains '+repr(nc)+' '+repr(sym)+' atoms\n')
                nc=len([atm for atm in solid2 if atm.symbol==sym])
                Optimizer.output.write('CX ROTCT_Defect: Defect 2 configuration contains '+repr(nc)+' '+repr(sym)+' atoms\n')
        if debug: Optimizer.output.flush()
        #pdb.set_trace()
        ind1[0]=indi1
        ind2[0]=indi2
    
    return ind1, ind2



import time
import os

def write_parameters(Optimizer):
    """
    Subprogram to write input parameters to output file
    Inputs:
        Optimizer = structopt Optimizer class object with output file data
    Outputs:
        None. Data is written directly to output files provided by Optimizer object
    """
    Optimizer.output.write('-----StructOpt - Global Structure Optimization Run----\n')
    if 'None' not in Optimizer.debug: Optimizer.output.write('*******DEBUGGING RUN*******\n')
    Optimizer.output.write(Optimizer.__version__ + '\n')
    localtime = time.asctime( time.localtime(time.time()) )
    Optimizer.output.write('Local time : ' + repr(localtime) + '\n')

    section_break = '\n--------------------\n'
    
    Optimizer.output.write(section_break + 'General Algorithm Information\n\n')
    if Optimizer.parallel:
        Optimizer.output.write('Algorithm Performed in parallel\n')
    else:
        Optimizer.output.write('Algorithm Performed in serial\n')
    Optimizer.output.write('Algorithm type : ' + Optimizer.algorithm_type + '\n')
    if 'Island_Method' in Optimizer.algorithm_type:
        Optimizer.output.write('    Migration Intervals : ' + repr(Optimizer.migration_intervals) + '\n')
        Optimizer.output.write('    Migration Percent : ' + repr(Optimizer.migration_percent) + '\n')
    Optimizer.output.write('Type of structure to optimize (structure) = ' + repr(Optimizer.structure) + '\n')
    Optimizer.output.write('Random number seed = ' + repr(Optimizer.seed) + '\n')
    if Optimizer.forcing=='Concentration':
        Optimizer.output.write('Concentration in individual held constant through crossover\n')
    elif Optimizer.forcing=='Energy_bias':
        Optimizer.output.write('Concentration in individual held constant through energy bias\n')
    elif Optimizer.forcing=='Chem_pot':
        Optimizer.output.write('Concentration in individual allowed to change based on Chemical Potentials\n')
    else:
        Optimizer.output.write('Concentration not held constant')
    Optimizer.output.write('Structure Fingerprinting (fingerprinting) : ' + repr(Optimizer.fingerprinting) + '\n')
    if Optimizer.fingerprinting:
        Optimizer.output.write('Fingerprint bin size (fpbin) : ' + repr(Optimizer.fpbin) + '\n')
        Optimizer.output.write('Fingerprint cutoff distance (fpcutoff) : ' + repr(Optimizer.fpcutoff) + '\n')
    if Optimizer.fixed_region: Optimizer.output.write('Fixed Bulk calculation \n')
    if Optimizer.constrain_position: Optimizer.output.write('Constrained position calculation \n')

    Optimizer.output.write(section_break + 'Population Generation Parameters\n\n')
    if Optimizer.restart: 
        Optimizer.output.write('Restarting Old Run\n')
        if Optimizer.restart_ints !=0: Optimizer.output.write('Assuming specified restart ints = ' + repr(Optimizer.restart_ints) + '\n')
    Optimizer.output.write('Number of structures in population (nindiv) : '  + repr(Optimizer.nindiv) + '\n')
    Optimizer.output.write('Starting structure configuration (atomlist) : '+ repr(Optimizer.atomlist) + '\n')
    Optimizer.output.write('Number of atoms in initial structure (natoms) : ' + repr(Optimizer.natoms) + '\n')
    Optimizer.output.write('Starting structure size (size) : ' + repr(Optimizer.size) + '\n')
    Optimizer.output.write('Average distance between atoms (r_ab) : ' + repr(Optimizer.r_ab) + '\n')
    Optimizer.output.write('Generation flag type = '+repr(Optimizer.generate_flag)+'\n')
    if Optimizer.structure=='Defect':
        Optimizer.output.write('File for bulk solid for Defect run (solidfile) : ' + repr(Optimizer.solidfile) + '\n')
        Optimizer.output.write('Cell size for bulk solid for Defect run (solidcell) : ' + repr(Optimizer.solidcell) + '\n')
        Optimizer.output.write('Size of bulk solid supercell (supercell) : ' + repr(Optimizer.supercell) + '\n')
        Optimizer.output.write('Evaluation of energy of bulk solid (evalsolid) : ' + repr(Optimizer.evalsolid) + '\n')
        Optimizer.output.write('Scale factor for surrounding material to include in Defect run (sf) : ' + repr(Optimizer.sf) + '\n')
        Optimizer.output.write('Start from random location option (Random_loc_start) : ' + repr(Optimizer.random_loc_start) + '\n')
        Optimizer.output.write('Start from random location option (random_vac_start) : ' + repr(Optimizer.random_vac_start) + '\n')
        Optimizer.output.write('Interstitial finding scheme is (finddefects) : ' + repr(Optimizer.finddefects) + '\n')
        Optimizer.output.write('Vacancy finding scheme is (trackvacs) : ' + repr(Optimizer.trackvacs) + '\n')
        Optimizer.output.write('Substitutions finding scheme is (trackswaps) : ' + repr(Optimizer.trackswaps) + '\n')
        if Optimizer.alloy:
            Optimizer.output.write('Allow free exchange between bulk and cluster (alloy) : ' + repr(Optimizer.alloy) + '\n')
    elif Optimizer.structure=='Crystal':
        Optimizer.output.write('Crystal cell shape options (cell_shape_options) : ' + repr(Optimizer.cell_shape_options) + '\n')

    Optimizer.output.write(section_break + 'Crossover Configuration Parameters\n\n')
    Optimizer.output.write('Crossover probability (cxpb) : ' + repr(Optimizer.cxpb) + '\n')
    Optimizer.output.write('Crossover scheme (cx_scheme) : ' + repr(Optimizer.cx_scheme) + '\n')
    Optimizer.output.write('Parent selection scheme (selection_scheme) : ' + repr(Optimizer.selection_scheme) + '\n')
    if 'tournament' in Optimizer.selection_scheme: Optimizer.output.write('    Tournament Size (tournsize) : ' + repr(Optimizer.tournsize) + '\n')
    if 'fuss' in Optimizer.selection_scheme:
        if Optimizer.selection_scheme != 'fuss1': 
            Optimizer.output.write('    fuss limit (fusslimit) : ' + repr(Optimizer.fusslimit) + ' eV \n')
        if Optimizer.natural_selection_scheme == 'fussf':
            Optimizer.output.write('Fingerprint bin size (fpbin) : ' + repr(Optimizer.fpbin) + '\n')
            Optimizer.output.write('Fingerprint cutoff distance (fpcutoff) : ' + repr(Optimizer.fpcutoff) + '\n')

    Optimizer.output.write(section_break + 'Mutation Configuration Parameters\n\n')
    Optimizer.output.write('Mutation Probability (mutpb) : ' + repr(Optimizer.mutpb) + '\n')
    Optimizer.output.write('Mutation Options (mutation_options) : ' + repr(Optimizer.mutation_options) + '\n')
    if Optimizer.swaplist==True or Optimizer.swaplist==None: Optimizer.output.write('Allow Constrained Swapping of atoms is True\n')
    BHFlag=False
    for one in Optimizer.mutation_options:
        if 'basin_hop' in one:
            BHFlag=True
    if BHFlag==True:
        Optimizer.output.write('Basin Hop mutation max steps (bh_steps) : ' + repr(Optimizer.bh_steps) + '\n')
        Optimizer.output.write('Basing Hop mutation temperature (bh_temp) : ' + repr(Optimizer.bh_temp) + '\n')
    if 'quench' in Optimizer.mutation_options:
        Optimizer.output.write('Quench mutation max temperature (K)' + repr(Optimizer.quench_max_temp)+'\n')
        Optimizer.output.write('Quench mutation min temperature (K)' +repr(Optimizer.quench_min_temp)+'\n')
        Optimizer.output.write('Quench mutation step size' +repr(Optimizer.quench_step_size)+'\n')
        Optimizer.output.write('Quench mutation number of steps 1'+repr(Optimizer.quench_n_steps_1)+'\n')
        Optimizer.output.write('Quench mutation number of steps 2'+repr(Optimizer.quench_n_steps_2)+'\n')
    if Optimizer.structure=='Defect':
        Optimizer.output.write('Setting mutation isolation to only region 1 = '+repr(Optimizer.isolate_mutation)+'\n')

    Optimizer.output.write(section_break + 'Individual Evaluation Parameters\n\n')
    if Optimizer.calc_method=='VASP':
        Optimizer.output.write('\n----VASP INPUT------\n')
        Optimizer.output.write(Optimizer.vaspcalc + '\n')
    elif Optimizer.calc_method=='LAMMPS':
        Optimizer.output.write('\n----LAMMPS Input------\n')
        Optimizer.output.write('The potential used is '+repr(Optimizer.pot_file)+'\n')
        Optimizer.output.write('The potential style is '+repr(Optimizer.pair_style)+'\n')
    elif Optimizer.calc_method=='Mixed':
        Optimizer.output.write('\n----Mixed Run------\n')
        if Optimizer.pot_file != None:
            Optimizer.output.write('The potential used is '+repr(Optimizer.pot_file)+'\n')
        if Optimizer.pair_style !=None:
            Optimizer.output.write('The potential style is ' + repr(Optimizer.pair_style) + '\n')
        else:
            Optimizer.output.write('A Leonard Jones Potential is assumed for LAMMPS\n')
        Optimizer.output.write('VASP calculations are performed with: \n' + Optimizer.vaspcalc + '\n')
    if Optimizer.ase_min==True:
        Optimizer.output.write('Using ASE BFGS Minimizer \n')
        Optimizer.output.write('fmax = ' + repr(Optimizer.ase_min_fmax) + '\n')
        Optimizer.output.write('Solver_Max_Steps = ' + repr(Optimizer.ase_min_maxsteps) + '\n')
    elif Optimizer.lammps_min != None:
        Optimizer.output.write('Lammps minimizer used ' + repr(Optimizer.lammps_min) + repr(Optimizer.lammps_min_style) + '\n')
    elif Optimizer.calc_method=='VASP':
        Optimizer.output.write('See VASP input to see local minimizer options\n')
    else:
        Optimizer.output.write('No Local Minimizer Used')
    if Optimizer.structure =='Cluster':
        Optimizer.output.write('LAMMPS box size for periodic evaluation = '+repr(Optimizer.large_box_size)+'\n')

    Optimizer.output.write(section_break+'Selection Parameters\n\n')
    Optimizer.output.write('Fitness Scheme (fitness_scheme) : ' + repr(Optimizer.fitness_scheme) + '\n')
    if 'stem' in Optimizer.fitness_scheme:
        Optimizer.output.write('Parameters for Convoluation AutoSTEM calculation:\n'+repr(Optimizer.stemcalc.parameters)+'\n')
        Optimizer.output.write('Keep STEM simulated images = ' + repr(Optimizer.stemcalc.keep_files) + '\n')
        Optimizer.output.write('STEM alpha coefficient = ' + repr(Optimizer.stem_coeff) + '\n')
    Optimizer.output.write('Natural Selection Scheme (natural_selection_scheme): ' + repr(Optimizer.natural_selection_scheme) + '\n')
    if 'tournament' in Optimizer.natural_selection_scheme: Optimizer.output.write('    Tournament Size (tournsize) = ' + repr(Optimizer.tournsize) + '\n')
    if 'fuss' in Optimizer.natural_selection_scheme:
        if Optimizer.natural_selection_scheme != 'fuss1': 
            Optimizer.output.write('FUSS limit (fusslim) : ' + repr(Optimizer.fusslimit) + ' eV \n')
        if Optimizer.natural_selection_scheme == 'fussf':
            Optimizer.output.write('Fingerprint bin size (fpbin) : ' + repr(Optimizer.fpbin) + '\n')
            Optimizer.output.write('Fingerprint cutoff distance (fpcutoff) : ' + repr(Optimizer.fpcutoff) + '\n')
    if 'metropolis' in Optimizer.natural_selection_scheme: Optimizer.output.write('Starting Temperature for Metropolis Selection = '+repr(Optimizer.metropolis_temp)+'kT\n')

    Optimizer.output.write(section_break+'Convergence Parameters\n\n')
    Optimizer.output.write('Population convergence scheme (CONVERGENCE_SCHEME) : ' + repr(Optimizer.convergence_scheme) + '\n')
    Optimizer.output.write('Maximum number of generation (maxgen) = ' + repr(Optimizer.maxgen) + '\n')
    if 'rep' in Optimizer.convergence_scheme:
        Optimizer.output.write('Required generation of repetition of energy for convergence (reqrep) = ' + repr(Optimizer.reqrep) + '\n')
        Optimizer.output.write('Energy cutoff for repetition (tolerance) : ' + repr(Optimizer.tolerance) + '\n')
    Optimizer.output.write('Duplicate structure convergence control scheme (predator) '+ repr(Optimizer.predator) + '\n')
    Optimizer.output.write('Minimum energy difference for duplicate consideration (demin) : ' + repr(Optimizer.demin) + '\n')
    if Optimizer.predator=='adapting':
        Optimizer.output.write('Fraction of repeated generations at which to begin adaptation (adaptbegin) : ' + repr(Optimizer.adaptbegin) + '\n')
        Optimizer.output.write('Multiplier for adaptation exponential (adaptmultiplier) : ' + repr(Optimizer.adaptmultiplier) + '\n')

    Optimizer.output.write(section_break+'Output and Post-Processing Parameters\n\n')
    Optimizer.output.write('Structure atoms filename (filename) : ' + os.getcwd() + '/' + Optimizer.filename + '\n')
    Optimizer.output.write('Summary filename : ' + repr(Optimizer.summary.name) + '\n')
    Optimizer.output.write('Format for summary file (output_format) : ' + repr(Optimizer.output_format) + '\n')
    if Optimizer.genealogy: Optimizer.output.write('Genealogy File name : ' + repr(Optimizer.Genealogyfile.name) + '\n')
    if Optimizer.allenergyfile: Optimizer.output.write('All Energy file written \n')
    if Optimizer.lattice_concentration: Optimizer.output.write('Lattice Concentration File written \n')
    if Optimizer.indiv_defect_write: Optimizer.output.write('Individual cluster files will also be written \n')
    Optimizer.output.write('Vacancies output to final structures : ' + repr(Optimizer.vacancy_output) + '\n')
    Optimizer.summary.write(Optimizer.filename+'\n')

    Optimizer.output.flush()
    return


from MAST.structopt.tools import get_best
from MAST.structopt.predator import mutation_dups
import random

def energy_cluster(pop, Optimizer):
    """Use a k-means clustering approach to identify individuals for new population
    """
    #Get coordinates for each individual
    pts = [ind.energy for ind in pop]
    #Identify spatial bounds
    mins = min(pts)
    maxs = max(pts)
    STR=''
    attemptcount=10
    passflag=False
    while attemptcount > 0:
        #Select random initial centroid locations
        cents = []
        while len(cents) < Optimizer.nindiv:
            a = random.choice(pts)
            if a not in cents:
                cents.append(a)
        #Assign individual to closest centroid
        clustlist = [[] for i in range(Optimizer.nindiv)]
        for one in pts:
            ds = []
            for i in range(len(cents)):
                d = abs(one - cents[i])
                ds.append([d,i])
            ds.sort()
            loc = ds[0]
            clustlist[loc[1]].append(one)
        try:
            count = 0
            while True:
                count += 1
                #Calculate the centroid of the new cluster
                centsnew = []
                for one in clustlist:
                    if len(one) > 0:
                        m = [1.0]*len(one)
                        #cop = numpy.dot(m, one) / sum(m)
                        cop = [sum([es for es in one]) / sum(m)]
                        centsnew.append(cop)
                    else:
                        centsnew.append(one)
                #Assign individuals to new centroids
                clustlistn=[[] for i in range(Optimizer.nindiv)]
                for one in pts:
                    ds=[]
                    for i in range(len(centsnew)):
                        d = abs(one - cents[i])
                        ds.append([d,i])
                    ds.sort()
                    loc = ds[0]
                    clustlistn[loc[1]].append(one)
                if clustlistn==clustlist:
                    clustlist = clustlistn
                    centroids = centsnew
                    attemptcount=-10
                    passflag=True
                    break
                else:
                    clustlist = clustlistn
        except:
            attemptcount-=1
            STR+='WARNING: PREDATOR: K-means cluster difficulty\n'
            print 'WARNING: PREDATOR: K-means cluster difficulty'
    if passflag==True:
        #Select one individual from each cluster
        news = []
        for one in clustlist:
            #Random
            news.append(random.choice(one))
            #Minimum energy
            #one.sort()
            #news.append(one[0])

        indices = []
        for one in news:
            for i in range(len(pts)):
                if pts[i]==one:
                    indices.append(i)
                    break
        npop = [pop[ind] for ind in indices]
    else:
        STR+='WARNING: PREDATOR: Population converging. Unable to find nindiv distinct clusters\n'
        npop,str1 = Mutation_Dups(pop,Optimizer)
        STR+=str1
        #npop=pop
        STR+='WARNING: PREDATOR: Chose new population based on Mutation-Dups predator\n'
    pop = get_best(npop,len(npop))
    return pop


##############################################################
# This code is part of the MAterials Simulation Toolkit (MAST)
# 
# Maintainer: Tam Mayeshiba
# Last updated: 2014-04-25
##############################################################
raise NotImplementedError("Comment out this line if you are sure you want to run this file! You may want to delete initialize.rst before running make latexpdf or make html")
import os
#  cmd = "sphinx-apidoc --separate --force -o ./source %s" % os.environ['MAST_INSTALL_PATH']
cmd = "sphinx-apidoc --separate -o ./source %s" % os.environ['MAST_INSTALL_PATH']
os.system(cmd)



import random
import copy

def cxtp(ind1, ind2, Optimizer):
    """ Simple Two point Crossover
    Maintains concentration
    Maintains total number of atoms
    Random clusters exchanged not necessarily geometry based
    """
    if 'CX' in Optimizer.debug:
        debug = True
    else:
        debug = False
    Optimizer.output.write('Simple Two Pt Cx between individual ' + 
            repr(ind1.index)+' and individual '+repr(ind2.index)+'\n')
    indi1=ind1[0]
    indi2=ind2[0]
    pos1=indi1.get_positions()
    pos2=indi2.get_positions()
    size = min(len(pos1), len(pos2))
    cxpoint1 = random.randint(1, size)
    cxpoint2 = random.randint(1, size - 1)
    if cxpoint2 >= cxpoint1:
        cxpoint2 += 1
    else:			# Swap the two cx points
        cxpoint1, cxpoint2 = cxpoint2, cxpoint1
    posi1=copy.deepcopy(pos1)
    posi2=copy.deepcopy(pos2)
    posi1[cxpoint1:cxpoint2]=pos2[cxpoint1:cxpoint2]
    posi2[cxpoint1:cxpoint2]=pos1[cxpoint1:cxpoint2]
    indi1.set_positions(posi1)
    indi2.set_positions(posi2)
    ind1[0]=indi1
    ind2[0]=indi2
    return ind1, ind2



import os
from ase import Atom, Atoms
from ase.io import read
from MAST.structopt.inp_out import read_xyz, write_xyz

def gen_solid(solidfile,solidcell,outfilename,calc=False,calcmeth=None):
    """Function to load a bulk solid from a file for use in Defect structure optimization
    Inputs:
        solidfile=String of filename to load
        solidcell=List/Matrix of cell parameters for ASE Atoms class
        outfilename=String of filename to write solid
        calc=False/calculator object for evaluating energy of solid
        calcmeth='VASP' or other method for calculating the energy of the solid
    Outputs:
        solid as ASE Atoms class
        energy and string if calc is not false
    """
    try:
        sol = read_xyz(solidfile)
    except Exception as e1:
        try:
            sol = read(solidfile)
        except Exception as e2:
            raise RuntimeError('Encountered errror:'+repr(e1)+' '+repr(e2)+
                ' While trying to read solid file given as:'+repr(solidfile))
    cell = solidcell
    sol.set_cell(cell)
    sol.set_pbc(True)
    #Evaluate pure Bulk structure
    if calc:
        cwd = os.getcwd()
        sol.set_calculator(calc)
        stro = ''
        try:
            if calcmeth == 'VASP':
                en = sol.get_potential_energy()
                calcb = Vasp(restart=True)
                sol = calcb.get_atoms()
                PureBulkEnpa = en/sol.get_number_of_atoms()
            else:
                OUT = sol.calc.calculate(sol)
                PureBulkEnpa = OUT['thermo'][-1]['pe']/sol.get_number_of_atoms()
                sol = OUT['atoms']
                sol.set_pbc(True)
        except:
            stro = 'WARNING: Unable to calculate energy of pure bulk solid'
            PureBulkEnpa = 0
        os.chdir(cwd)
        # Write bulk file to directory
        write_xyz(outfilename,sol,PureBulkEnpa)
        return sol, PureBulkEnpa, stro
    else:
        # Write bulk file to directory
        write_xyz(outfilename,sol,'Pure Bulk')
        return sol